import asyncio
from typing import Optional, Literal
from unittest import mock
import uuid

import aiohttp
import pytest
from aiohttp.client_reqrep import RequestInfo
from multidict import CIMultiDict, CIMultiDictProxy
from pydantic import BaseModel
from yarl import URL

from meshagent.agents.adapter import LLMAdapter
from meshagent.agents.chat import ChatBot, ChatBotClient, ChatThreadContext
from meshagent.agents.context import AgentSessionContext
from meshagent.agents.thread_schema import thread_list_schema, thread_schema
from meshagent.api import RoomException
from meshagent.api.chan import Chan
from meshagent.api.messaging import JsonContent
from meshagent.api.participant import Participant
from meshagent.openai.tools.responses_adapter import MCPToolkitBuilder
from meshagent.tools import ToolContext, ToolkitBuilder, Toolkit


class _FakeStorage:
    def __init__(self, *, existing_paths: Optional[set[str]] = None):
        self._existing_paths = existing_paths or set()

    async def exists(self, *, path: str) -> bool:
        return path in self._existing_paths

    async def list(self, *, path: str):
        del path
        return []


class _FakeMessaging:
    def __init__(self):
        self.sent_messages: list[dict] = []
        self.handlers: dict[str, object] = {}
        self.enabled = False

    def get_participants(self):
        return []

    def on(self, event: str, handler) -> None:
        self.handlers[event] = handler

    def off(self, event: str, handler) -> None:
        existing = self.handlers.get(event)
        if existing is handler:
            self.handlers.pop(event, None)

    async def enable(self) -> None:
        self.enabled = True

    def send_message_nowait(self, *, to, type, message):
        self.sent_messages.append({"to": to, "type": type, "message": message})

    async def send_message(
        self,
        *,
        to,
        type,
        message,
        attachment=None,
    ) -> None:
        del attachment
        self.send_message_nowait(to=to, type=type, message=message)


class _FakeAgents:
    async def list_toolkits(
        self,
        *,
        participant_id: Optional[str] = None,
        participant_name: Optional[str] = None,
    ):
        del participant_id
        del participant_name
        return []


class _FakeThreadListElement:
    def __init__(self, *, tag_name: str, attributes: dict[str, str]):
        self.tag_name = tag_name
        self._attributes = dict(attributes)

    def get_attribute(self, name: str):
        return self._attributes.get(name)

    def set_attribute(self, name: str, value):
        self._attributes[name] = value


class _FakeThreadElement:
    def __init__(
        self,
        *,
        tag_name: str,
        attributes: Optional[dict[str, object]] = None,
    ):
        self.tag_name = tag_name
        self._attributes = dict(attributes or {})
        self._children: list["_FakeThreadElement"] = []

    def get_attribute(self, name: str):
        return self._attributes.get(name)

    def set_attribute(self, name: str, value) -> None:
        self._attributes[name] = value

    def get_children(self) -> list["_FakeThreadElement"]:
        return [*self._children]

    def get_children_by_tag_name(self, tag_name: str) -> list["_FakeThreadElement"]:
        return [child for child in self._children if child.tag_name == tag_name]

    def append_child(
        self,
        *,
        tag_name: str,
        attributes: Optional[dict[str, object]] = None,
    ) -> "_FakeThreadElement":
        element = _FakeThreadElement(tag_name=tag_name, attributes=attributes)
        self._children.append(element)
        return element


class _FakeThreadDocument:
    def __init__(self):
        self.root = _FakeThreadElement(tag_name="thread")


class _FakeThreadListRoot:
    def __init__(self):
        self._children: list[_FakeThreadListElement] = []

    def get_children(self) -> list[_FakeThreadListElement]:
        return [*self._children]

    def append_child(
        self, *, tag_name: str, attributes: dict[str, str]
    ) -> _FakeThreadListElement:
        element = _FakeThreadListElement(tag_name=tag_name, attributes=attributes)
        self._children.append(element)
        return element


class _FakeThreadListDocument:
    def __init__(self):
        self.root = _FakeThreadListRoot()


class _FakeSync:
    def __init__(self, *, document: Optional[_FakeThreadListDocument] = None):
        self.document = document or _FakeThreadListDocument()
        self.open_calls: list[dict] = []
        self.close_calls: list[str] = []

    async def open(
        self,
        *,
        path: str,
        create: bool = True,
        initial_json: Optional[dict] = None,
        schema=None,
    ):
        del create
        del initial_json
        self.open_calls.append({"path": path, "schema": schema})
        return self.document

    async def close(self, *, path: str) -> None:
        self.close_calls.append(path)


class _FakeLocalParticipant(Participant):
    def __init__(self):
        super().__init__(
            id="assistant-id",
            attributes={"name": "assistant"},
        )
        self.set_attribute_calls: list[tuple[str, object]] = []

    async def set_attribute(self, name: str, value) -> None:
        self._attributes[name] = value
        self.set_attribute_calls.append((name, value))


class _FakeRoom:
    def __init__(
        self,
        *,
        existing_paths: Optional[set[str]] = None,
        sync: Optional[_FakeSync] = None,
    ):
        self.local_participant = _FakeLocalParticipant()
        self.storage = _FakeStorage(existing_paths=existing_paths)
        self.messaging = _FakeMessaging()
        self.sync = sync or _FakeSync()
        self.agents = _FakeAgents()


class _FakeQueue:
    def __init__(self):
        self.items = []

    def send_nowait(self, item) -> None:
        self.items.append(item)


class _ExplodingMessageChannel:
    def __init__(self, *, error: Exception):
        self._error = error

    async def recv(self):
        raise self._error


class _BlockingMessageChannel:
    async def recv(self):
        await asyncio.Future()


class _TrackedThreadContext:
    def __init__(self, *, path: str):
        self.path = path
        self.closed = False

    async def __aexit__(self, exc_type, exc, tb) -> None:
        del exc_type
        del exc
        del tb
        self.closed = True


def _make_ws_handshake_error(
    *, status: int, headers: dict[str, str] | None = None
) -> aiohttp.WSServerHandshakeError:
    request_headers = CIMultiDictProxy(CIMultiDict())
    url = URL("ws://localhost:8080/openai/v1/responses")
    return aiohttp.WSServerHandshakeError(
        request_info=RequestInfo(
            url=url,
            method="GET",
            headers=request_headers,
            real_url=url,
        ),
        history=(),
        status=status,
        message="Invalid response status",
        headers=CIMultiDictProxy(CIMultiDict(headers or {})),
    )


class _FakeOpenThreadAdapter:
    def make_toolkit(self) -> Toolkit:
        return Toolkit(name="open thread", tools=[])


class _FakeThreadNameAdapter(LLMAdapter):
    def __init__(self, *, generated_thread_name: str):
        self.generated_thread_name = generated_thread_name
        self.calls: list[dict] = []

    def default_model(self) -> str:
        return "thread-name-model"

    async def next(
        self,
        *,
        context,
        room,
        toolkits,
        output_schema=None,
        event_handler=None,
        steering_callback=None,
        model=None,
        on_behalf_of=None,
        options: Optional[dict] = None,
    ):
        del context
        del room
        del toolkits
        del event_handler
        del steering_callback
        del on_behalf_of
        del options
        self.calls.append({"output_schema": output_schema, "model": model})
        return {"thread_name": self.generated_thread_name}


class _SessionRequiredContext(AgentSessionContext):
    pass


class _SessionRequiredThreadNameAdapter(LLMAdapter):
    def __init__(self, *, generated_thread_name: str):
        self.generated_thread_name = generated_thread_name
        self.last_messages: list[dict] = []
        self.last_context_type: type[AgentSessionContext] | None = None

    def default_model(self) -> str:
        return "thread-name-model"

    def create_session(self) -> AgentSessionContext:
        return _SessionRequiredContext(system_role=None)

    async def next(
        self,
        *,
        context,
        room,
        toolkits,
        output_schema=None,
        event_handler=None,
        steering_callback=None,
        model=None,
        on_behalf_of=None,
        options: Optional[dict] = None,
    ):
        del room
        del toolkits
        del output_schema
        del event_handler
        del steering_callback
        del model
        del on_behalf_of
        del options
        self.last_context_type = type(context)
        self.last_messages = [*context.messages]
        if not isinstance(context, _SessionRequiredContext):
            raise RuntimeError("expected adapter-created session context")
        return {"thread_name": self.generated_thread_name}


class _CaptureChatAdapter(LLMAdapter):
    def __init__(self):
        self.last_messages: list[dict] = []

    def default_model(self) -> str:
        return "chat-model"

    async def next(
        self,
        *,
        context,
        room,
        toolkits,
        output_schema=None,
        event_handler=None,
        steering_callback=None,
        model=None,
        on_behalf_of=None,
        options: Optional[dict] = None,
    ):
        del room
        del toolkits
        del output_schema
        del event_handler
        del steering_callback
        del model
        del on_behalf_of
        del options
        self.last_messages = [*context.messages]
        return "ok"


class _ShouldReplyAdapter(LLMAdapter):
    def __init__(self, *, response):
        self.response = response

    def default_model(self) -> str:
        return "decision-model"

    async def next(
        self,
        *,
        context,
        room,
        toolkits,
        output_schema=None,
        event_handler=None,
        steering_callback=None,
        model=None,
        on_behalf_of=None,
        options: Optional[dict] = None,
    ):
        del context
        del room
        del toolkits
        del output_schema
        del event_handler
        del steering_callback
        del model
        del on_behalf_of
        del options
        return self.response


class _ExampleToolkitConfig(BaseModel):
    name: Literal["example"] = "example"
    enabled: bool = False


class _ExampleToolkitBuilder(ToolkitBuilder):
    def __init__(self):
        super().__init__(name="example", type=_ExampleToolkitConfig)

    async def make(self, *, room, model: str, config: _ExampleToolkitConfig) -> Toolkit:
        del room
        del model
        del config
        return Toolkit(name="example", tools=[])


class _ChatBotWithToolBuilders(ChatBot):
    def get_toolkit_builders(self) -> list[ToolkitBuilder]:
        return [_ExampleToolkitBuilder()]


class _ChatBotWithMCPToolkitBuilder(ChatBot):
    def get_toolkit_builders(self) -> list[ToolkitBuilder]:
        return [MCPToolkitBuilder()]


class _ChatBotAlwaysReplies(ChatBot):
    async def should_reply(
        self,
        *,
        has_more_than_one_other_user: bool,
        online: list[Participant],
        context: ChatThreadContext,
        toolkits: list[Toolkit],
        from_user: Participant,
    ) -> bool:
        del has_more_than_one_other_user
        del online
        del context
        del toolkits
        del from_user
        return True


@pytest.mark.asyncio
async def test_chatbot_client_send_requests_server_side_store_without_touching_doc():
    class _ExplodingDocument:
        @property
        def root(self):
            raise AssertionError("chat client should not write directly to the thread")

    room = _FakeRoom()
    client = ChatBotClient(
        room=room,
        participant_name="assistant",
        thread_path="/threads/test.thread",
    )
    client._participant = Participant(  # type: ignore[assignment]
        id="assistant-id",
        attributes={"name": "assistant"},
    )
    client._doc = _ExplodingDocument()

    await client.send(
        text="hello",
        attachments=[{"path": "uploads/report.pdf"}],
    )

    assert room.messaging.sent_messages == [
        {
            "to": client._participant,
            "type": "chat",
            "message": {
                "text": "hello",
                "path": "/threads/test.thread",
                "tools": [],
                "attachments": [{"path": "uploads/report.pdf"}],
                "store": True,
            },
        }
    ]


@pytest.mark.asyncio
async def test_send_and_save_chat_uses_thread_adapter_write_text_message() -> None:
    class _RecordingThreadAdapter:
        def __init__(self) -> None:
            self.calls: list[dict[str, object]] = []
            self.path = "/threads/test.thread"

        def write_text_message(
            self,
            *,
            text: str,
            participant,
            attachments=None,
        ) -> None:
            self.calls.append(
                {
                    "text": text,
                    "participant": participant,
                    "attachments": attachments,
                }
            )

    room = _FakeRoom()
    bot = ChatBot(llm_adapter=_CaptureChatAdapter())
    bot._room = room
    adapter = _RecordingThreadAdapter()
    recipient = Participant(id="caller-id", attributes={"name": "alice"})

    await bot._send_and_save_chat(
        thread_adapter=adapter,  # type: ignore[arg-type]
        to=recipient,  # type: ignore[arg-type]
        id="message-1",
        text="hello",
        thread_attributes={"path": "/threads/test.thread"},
    )

    assert room.messaging.sent_messages == [
        {
            "to": recipient,
            "type": "chat",
            "message": {"path": "/threads/test.thread", "text": "hello"},
        }
    ]
    assert adapter.calls == [
        {
            "text": "hello",
            "participant": room.local_participant,
            "attachments": None,
        }
    ]

    async def get_thread_toolkits(
        self, *, thread_context: ChatThreadContext, participant: Participant
    ) -> list[Toolkit]:
        del thread_context
        del participant
        return []


async def _new_thread_tool(bot: ChatBot):
    toolkits = await bot.get_exposed_toolkits()
    chatbot_toolkit = next(
        toolkit
        for toolkit in toolkits
        if any(tool.name == "new_thread" for tool in toolkit.tools)
    )
    return next(tool for tool in chatbot_toolkit.tools if tool.name == "new_thread")


def _stub_new_thread_member_seed(bot: ChatBot) -> mock.AsyncMock:
    stub = mock.AsyncMock()
    bot._seed_new_thread_members = stub  # type: ignore[method-assign]
    return stub


async def _chat_toolkit(bot: ChatBot):
    toolkits = await bot.get_exposed_toolkits()
    return next(toolkit for toolkit in toolkits if toolkit.name == "chat")


def _assert_uuid_thread_path(*, path: str, prefix: str) -> None:
    assert path.startswith(prefix)
    assert path.endswith(".thread")
    basename = path[len(prefix) : -len(".thread")]
    parsed = uuid.UUID(basename)
    assert str(parsed) == basename


@pytest.mark.asyncio
async def test_new_thread_tool_creates_named_thread_and_queues_message() -> None:
    adapter = _FakeThreadNameAdapter(generated_thread_name="Release Planning / Q1")
    bot = ChatBot(llm_adapter=adapter)
    room = _FakeRoom()
    bot._room = room
    seeded_members = _stub_new_thread_member_seed(bot)

    queue = _FakeQueue()

    def _ensure_thread(*, path: str):
        del path
        return queue

    bot._ensure_thread = _ensure_thread  # type: ignore[method-assign]

    tool = await _new_thread_tool(bot)
    tools_schema = tool.input_schema["properties"]["message"]["properties"]["tools"]
    assert len(tools_schema["anyOf"]) == 2
    assert tools_schema["anyOf"][0]["type"] == "array"
    assert tools_schema["anyOf"][0]["items"]["type"] == "object"
    assert tools_schema["anyOf"][0]["items"]["additionalProperties"] is False
    assert tools_schema["anyOf"][0]["items"]["properties"] == {}
    assert tools_schema["anyOf"][0]["items"]["required"] == []
    assert tools_schema["anyOf"][1] == {"type": "null"}
    context = ToolContext(
        room=room,
        caller=Participant(id="caller-id", attributes={"name": "alice"}),
    )
    result = await tool.execute(
        context=context,
        message={
            "text": "Plan the Q1 release milestones",
            "attachments": [{"path": "uploads/plan.md"}],
        },
    )

    assert isinstance(result, JsonContent)
    result_path = result.json["path"]
    assert result.json["name"] == "Release Planning Q1"
    _assert_uuid_thread_path(path=result_path, prefix=".threads/assistant/")

    assert len(queue.items) == 1
    queued = queue.items[0]
    assert queued.type == "chat"
    assert queued.message["path"] == result_path
    assert queued.message["text"] == "Plan the Q1 release milestones"
    assert queued.message["attachments"] == [{"path": "uploads/plan.md"}]
    assert queued.message["store"] is True
    assert not queued.result.done()

    seeded_members.assert_awaited_once()
    assert seeded_members.await_args.kwargs["path"] == result_path
    assert seeded_members.await_args.kwargs["members"][0] is room.local_participant
    assert seeded_members.await_args.kwargs["members"][1] is context.caller

    assert len(adapter.calls) == 1
    assert adapter.calls[0]["output_schema"] is not None


@pytest.mark.asyncio
async def test_new_thread_tool_uses_thread_dir_for_guid_path() -> None:
    adapter = _FakeThreadNameAdapter(generated_thread_name="Release Planning")
    bot = ChatBot(llm_adapter=adapter, thread_dir="custom")
    room = _FakeRoom()
    bot._room = room
    _stub_new_thread_member_seed(bot)

    queue = _FakeQueue()

    def _ensure_thread(*, path: str):
        del path
        return queue

    bot._ensure_thread = _ensure_thread  # type: ignore[method-assign]

    tool = await _new_thread_tool(bot)
    tools_schema = tool.input_schema["properties"]["message"]["properties"]["tools"]
    assert len(tools_schema["anyOf"]) == 2
    assert tools_schema["anyOf"][0]["type"] == "array"
    assert tools_schema["anyOf"][0]["items"]["type"] == "object"
    assert tools_schema["anyOf"][0]["items"]["additionalProperties"] is False
    assert tools_schema["anyOf"][0]["items"]["properties"] == {}
    assert tools_schema["anyOf"][0]["items"]["required"] == []
    assert tools_schema["anyOf"][1] == {"type": "null"}
    context = ToolContext(
        room=room,
        caller=Participant(id="caller-id", attributes={"name": "alice"}),
    )
    result = await tool.execute(
        context=context,
        message={"text": "Plan the release"},
    )

    assert isinstance(result, JsonContent)
    result_path = result.json["path"]
    _assert_uuid_thread_path(path=result_path, prefix="custom/")
    assert len(queue.items) == 1
    queued = queue.items[0]
    assert queued.message["path"] == result_path
    assert queued.message["text"] == "Plan the release"
    assert queued.message["attachments"] == []
    assert queued.message["store"] is True


@pytest.mark.asyncio
async def test_new_thread_tool_accepts_empty_tools_without_builder_schema() -> None:
    adapter = _FakeThreadNameAdapter(generated_thread_name="No Builder Tools")
    bot = ChatBot(llm_adapter=adapter)
    room = _FakeRoom()
    bot._room = room
    _stub_new_thread_member_seed(bot)

    queue = _FakeQueue()

    def _ensure_thread(*, path: str):
        del path
        return queue

    bot._ensure_thread = _ensure_thread  # type: ignore[method-assign]

    tool = await _new_thread_tool(bot)
    tools_schema = tool.input_schema["properties"]["message"]["properties"]["tools"]
    assert len(tools_schema["anyOf"]) == 2
    assert tools_schema["anyOf"][0]["type"] == "array"
    assert tools_schema["anyOf"][0]["items"]["type"] == "object"
    assert tools_schema["anyOf"][0]["items"]["additionalProperties"] is False
    assert tools_schema["anyOf"][0]["items"]["properties"] == {}
    assert tools_schema["anyOf"][0]["items"]["required"] == []
    assert tools_schema["anyOf"][1] == {"type": "null"}
    context = ToolContext(
        room=room,
        caller=Participant(id="caller-id", attributes={"name": "alice"}),
    )
    result = await tool.execute(
        context=context,
        message={
            "text": "Thread with empty tools",
            "attachments": [],
            "tools": [],
        },
    )

    assert isinstance(result, JsonContent)
    _assert_uuid_thread_path(
        path=result.json["path"],
        prefix=".threads/assistant/",
    )
    assert result.json["name"] == "No Builder Tools"
    assert len(queue.items) == 1
    queued = queue.items[0]
    assert queued.message["tools"] == []
    assert queued.message["store"] is True


@pytest.mark.asyncio
async def test_new_thread_tool_schema_supports_mcp_toolkit_builder() -> None:
    adapter = _FakeThreadNameAdapter(generated_thread_name="MCP")
    bot = _ChatBotWithMCPToolkitBuilder(llm_adapter=adapter)
    room = _FakeRoom()
    bot._room = room

    tool = await _new_thread_tool(bot)
    tools_schema = tool.input_schema["properties"]["message"]["properties"]["tools"]
    items_schema = tools_schema["anyOf"][0]["items"]
    headers_schema = items_schema["$defs"]["MCPServer"]["properties"]["headers"][
        "anyOf"
    ][0]

    assert headers_schema["type"] == "array"
    assert headers_schema["items"]["$ref"] == "#/$defs/Header"
    assert items_schema["$defs"]["Header"]["additionalProperties"] is False


@pytest.mark.asyncio
async def test_new_thread_tool_accepts_tools_with_toolkit_builder_schema() -> None:
    adapter = _FakeThreadNameAdapter(generated_thread_name="Builder Thread")
    bot = _ChatBotWithToolBuilders(llm_adapter=adapter)
    room = _FakeRoom()
    bot._room = room
    _stub_new_thread_member_seed(bot)

    queue = _FakeQueue()

    def _ensure_thread(*, path: str):
        del path
        return queue

    bot._ensure_thread = _ensure_thread  # type: ignore[method-assign]

    tool = await _new_thread_tool(bot)
    tools_schema = tool.input_schema["properties"]["message"]["properties"]["tools"]
    assert len(tools_schema["anyOf"]) == 2
    assert tools_schema["anyOf"][0]["type"] == "array"
    assert tools_schema["anyOf"][0]["items"]["type"] == "object"
    assert tools_schema["anyOf"][0]["items"]["properties"]["name"]["type"] == "string"
    assert tools_schema["anyOf"][1] == {"type": "null"}
    context = ToolContext(
        room=room,
        caller=Participant(id="caller-id", attributes={"name": "alice"}),
    )
    result = await tool.execute(
        context=context,
        message={
            "text": "Thread with tool config",
            "tools": [{"name": "example", "enabled": True}],
        },
    )

    assert isinstance(result, JsonContent)
    _assert_uuid_thread_path(
        path=result.json["path"],
        prefix=".threads/assistant/",
    )
    assert result.json["name"] == "Builder Thread"
    assert len(queue.items) == 1
    queued = queue.items[0]
    assert queued.message["tools"] == [{"name": "example", "enabled": True}]
    assert queued.message["store"] is True


@pytest.mark.asyncio
async def test_new_thread_tool_uses_adapter_created_context_for_thread_naming() -> None:
    adapter = _SessionRequiredThreadNameAdapter(generated_thread_name="Adapter Context")
    bot = ChatBot(llm_adapter=adapter)
    room = _FakeRoom()
    bot._room = room
    _stub_new_thread_member_seed(bot)

    queue = _FakeQueue()

    def _ensure_thread(*, path: str):
        del path
        return queue

    bot._ensure_thread = _ensure_thread  # type: ignore[method-assign]

    tool = await _new_thread_tool(bot)
    context = ToolContext(
        room=room,
        caller=Participant(id="caller-id", attributes={"name": "alice"}),
        caller_context={
            "chat": {
                "messages": [{"role": "user", "content": "prior context"}],
                "system_role": None,
                "previous_messages": [],
                "previous_response_id": None,
            }
        },
    )
    result = await tool.execute(
        context=context,
        message={"text": "Name this from adapter context"},
    )

    assert isinstance(result, JsonContent)
    _assert_uuid_thread_path(
        path=result.json["path"],
        prefix=".threads/assistant/",
    )
    assert result.json["name"] == "Adapter Context"
    assert adapter.last_context_type is _SessionRequiredContext
    assert [m.get("content") for m in adapter.last_messages] == [
        "prior context",
        "Name this from adapter context",
    ]


@pytest.mark.asyncio
async def test_new_thread_tool_records_friendly_name_in_thread_list() -> None:
    adapter = _FakeThreadNameAdapter(generated_thread_name="Friendly Plan Name")
    bot = ChatBot(llm_adapter=adapter, thread_dir="custom")
    room = _FakeRoom()
    bot._room = room
    _stub_new_thread_member_seed(bot)

    doc = _FakeThreadListDocument()
    bot._thread_list_document = doc
    bot._thread_list_path = "custom/index.threadl"

    queue = _FakeQueue()

    def _ensure_thread(*, path: str):
        del path
        return queue

    bot._ensure_thread = _ensure_thread  # type: ignore[method-assign]

    tool = await _new_thread_tool(bot)
    context = ToolContext(
        room=room,
        caller=Participant(id="caller-id", attributes={"name": "alice"}),
    )
    result = await tool.execute(
        context=context,
        message={"text": "Plan this friendly thread"},
    )

    path = result.json["path"]
    assert result.json["name"] == "Friendly Plan Name"
    _assert_uuid_thread_path(path=path, prefix="custom/")
    entries = doc.root.get_children()
    assert len(entries) == 1
    assert entries[0].get_attribute("name") == "Friendly Plan Name"
    assert entries[0].get_attribute("path") == path


@pytest.mark.asyncio
async def test_seed_new_thread_members_adds_agent_and_sender_without_duplicates() -> (
    None
):
    adapter = _CaptureChatAdapter()
    bot = ChatBot(llm_adapter=adapter)
    thread = _FakeThreadDocument()
    members = thread.root.append_child(tag_name="members")
    members.append_child(tag_name="member", attributes={"name": "assistant"})

    sync = mock.Mock()
    sync.open = mock.AsyncMock(return_value=thread)
    sync.close = mock.AsyncMock()

    room = _FakeRoom(sync=sync)
    bot._room = room

    sender = Participant(id="caller-id", attributes={"name": "alice"})

    await bot._seed_new_thread_members(
        path=".threads/assistant/test.thread",
        members=[room.local_participant, sender, sender],
    )

    assert sync.open.await_args.kwargs == {
        "path": ".threads/assistant/test.thread",
        "schema": thread_schema,
    }
    sync.close.assert_awaited_once_with(path=".threads/assistant/test.thread")
    assert [child.get_attribute("name") for child in members.get_children()] == [
        "assistant",
        "alice",
    ]


@pytest.mark.asyncio
async def test_thread_list_tools_only_present_when_thread_dir_is_set() -> None:
    adapter = _CaptureChatAdapter()

    bot_without_thread_dir = ChatBot(llm_adapter=adapter)
    room_without_thread_dir = _FakeRoom()
    bot_without_thread_dir._room = room_without_thread_dir
    toolkit_without_thread_dir = await _chat_toolkit(bot_without_thread_dir)
    names_without = {tool.name for tool in toolkit_without_thread_dir.tools}
    assert "list_threads" not in names_without
    assert "grep_thread_list" not in names_without

    bot_with_thread_dir = ChatBot(llm_adapter=adapter, thread_dir="custom")
    room_with_thread_dir = _FakeRoom()
    bot_with_thread_dir._room = room_with_thread_dir
    toolkit_with_thread_dir = await _chat_toolkit(bot_with_thread_dir)
    names_with = {tool.name for tool in toolkit_with_thread_dir.tools}
    assert "list_threads" in names_with
    assert "grep_thread_list" in names_with


@pytest.mark.asyncio
async def test_default_new_threading_mode_enables_thread_list_tools_and_relaxed_remote_validation() -> (
    None
):
    adapter = _CaptureChatAdapter()
    sync = _FakeSync()
    room = _FakeRoom(sync=sync)
    bot = ChatBot(llm_adapter=adapter, threading_mode="default-new")
    bot._room = room

    toolkit = await _chat_toolkit(bot)
    tool_names = {tool.name for tool in toolkit.tools}

    assert toolkit.validation_mode == "content_types"
    assert "list_threads" in tool_names
    assert "grep_thread_list" in tool_names

    await bot._open_thread_list_document()

    assert sync.open_calls == [
        {
            "path": ".threads/assistant/index.threadl",
            "schema": thread_list_schema,
        }
    ]


@pytest.mark.asyncio
async def test_get_thread_toolkits_includes_thread_list_tools_only() -> None:
    adapter = _CaptureChatAdapter()
    bot = ChatBot(llm_adapter=adapter, thread_dir="custom")
    room = _FakeRoom()
    bot._room = room

    path = "custom/main.thread"
    bot._open_threads[path] = _FakeOpenThreadAdapter()

    thread = mock.Mock()
    thread.root = mock.Mock()
    thread.root.get_elements_by_tag_name.return_value = [mock.Mock()]

    thread_context = ChatThreadContext(
        session=AgentSessionContext(),
        thread=thread,
        path=path,
    )
    participant = Participant(id="caller-id", attributes={"name": "alice"})

    toolkits = await bot.get_thread_toolkits(
        thread_context=thread_context,
        participant=participant,
    )
    tool_names = {tool.name for toolkit in toolkits for tool in toolkit.tools}

    assert "list_threads" in tool_names
    assert "grep_thread_list" in tool_names
    assert "ask" not in tool_names
    assert "new_thread" not in tool_names


@pytest.mark.asyncio
async def test_get_thread_toolkits_omits_thread_list_tools_without_thread_dir() -> None:
    adapter = _CaptureChatAdapter()
    bot = ChatBot(llm_adapter=adapter)
    room = _FakeRoom()
    bot._room = room

    path = ".threads/assistant/main.thread"
    bot._open_threads[path] = _FakeOpenThreadAdapter()

    thread = mock.Mock()
    thread.root = mock.Mock()
    thread.root.get_elements_by_tag_name.return_value = [mock.Mock()]

    thread_context = ChatThreadContext(
        session=AgentSessionContext(),
        thread=thread,
        path=path,
    )
    participant = Participant(id="caller-id", attributes={"name": "alice"})

    toolkits = await bot.get_thread_toolkits(
        thread_context=thread_context,
        participant=participant,
    )
    tool_names = {tool.name for toolkit in toolkits for tool in toolkit.tools}

    assert "list_threads" not in tool_names
    assert "grep_thread_list" not in tool_names


@pytest.mark.asyncio
async def test_list_threads_sorts_by_modified_desc_and_supports_offset_limit() -> None:
    adapter = _CaptureChatAdapter()
    bot = ChatBot(llm_adapter=adapter, thread_dir="custom")
    room = _FakeRoom()
    bot._room = room

    doc = _FakeThreadListDocument()
    doc.root.append_child(
        tag_name="thread",
        attributes={
            "name": "first",
            "path": "custom/a.thread",
            "created_at": "2024-01-01T00:00:00Z",
            "modified_at": "2024-01-01T00:00:00Z",
        },
    )
    doc.root.append_child(
        tag_name="thread",
        attributes={
            "name": "second",
            "path": "custom/b.thread",
            "created_at": "2024-01-01T00:00:00Z",
            "modified_at": "2025-01-01T00:00:00Z",
        },
    )
    doc.root.append_child(
        tag_name="thread",
        attributes={
            "name": "third",
            "path": "custom/c.thread",
            "created_at": "2024-01-01T00:00:00Z",
            "modified_at": "2023-01-01T00:00:00Z",
        },
    )
    bot._thread_list_document = doc

    toolkit = await _chat_toolkit(bot)
    list_tool = next(tool for tool in toolkit.tools if tool.name == "list_threads")
    context = ToolContext(
        room=room,
        caller=Participant(id="caller-id", attributes={"name": "alice"}),
    )

    result = await list_tool.execute(context=context, limit=2, offset=1)
    assert isinstance(result, JsonContent)
    assert result.json["sort"] == "modified_at_desc"
    assert result.json["total"] == 3
    assert result.json["offset"] == 1
    assert result.json["limit"] == 2
    names = [entry["name"] for entry in result.json["threads"]]
    assert names == ["first", "third"]
    assert "Use read_file with a thread path" in result.json["read_file_hint"]


@pytest.mark.asyncio
async def test_grep_thread_list_finds_matches_and_mentions_read_file() -> None:
    adapter = _CaptureChatAdapter()
    bot = ChatBot(llm_adapter=adapter, thread_dir="custom")
    room = _FakeRoom()
    bot._room = room

    doc = _FakeThreadListDocument()
    doc.root.append_child(
        tag_name="thread",
        attributes={
            "name": "Release Plan",
            "path": "custom/release.thread",
            "created_at": "2024-01-01T00:00:00Z",
            "modified_at": "2025-01-01T00:00:00Z",
        },
    )
    doc.root.append_child(
        tag_name="thread",
        attributes={
            "name": "Unrelated",
            "path": "custom/other.thread",
            "created_at": "2024-01-01T00:00:00Z",
            "modified_at": "2024-01-02T00:00:00Z",
        },
    )
    bot._thread_list_document = doc

    toolkit = await _chat_toolkit(bot)
    grep_tool = next(tool for tool in toolkit.tools if tool.name == "grep_thread_list")
    context = ToolContext(
        room=room,
        caller=Participant(id="caller-id", attributes={"name": "alice"}),
    )

    result = await grep_tool.execute(
        context=context, pattern="release", ignore_case=True
    )
    assert isinstance(result, JsonContent)
    assert result.json["total_matches"] == 1
    assert result.json["pattern"] == "release"
    assert result.json["ignore_case"] is True
    assert [entry["name"] for entry in result.json["threads"]] == ["Release Plan"]
    assert "Use read_file with a thread path" in result.json["read_file_hint"]


@pytest.mark.asyncio
async def test_on_chat_received_adds_current_file_context_message() -> None:
    adapter = _CaptureChatAdapter()
    bot = _ChatBotAlwaysReplies(llm_adapter=adapter)
    room = _FakeRoom()
    bot._room = room
    bot._open_threads[".threads/main.thread"] = _FakeOpenThreadAdapter()

    thread = mock.Mock()
    thread.root = mock.Mock()
    thread.root.get_children.return_value = []

    thread_context = ChatThreadContext(
        session=AgentSessionContext(),
        thread=thread,
        path=".threads/main.thread",
    )
    from_participant = Participant(
        id="caller-id",
        attributes={"name": "alice", "current_file": "docs/plan.md"},
    )

    result = await bot.on_chat_received(
        thread_context=thread_context,
        from_participant=from_participant,
        message={"text": "Summarize this"},
    )

    assert result == "ok"
    assert any(
        message.get("role") == "assistant"
        and message.get("content")
        == "alice is currently viewing the file at the path: docs/plan.md"
        for message in adapter.last_messages
    )


@pytest.mark.asyncio
async def test_on_chat_received_adds_not_viewing_message_when_file_is_closed() -> None:
    adapter = _CaptureChatAdapter()
    bot = _ChatBotAlwaysReplies(llm_adapter=adapter)
    room = _FakeRoom()
    bot._room = room
    bot._open_threads[".threads/main.thread"] = _FakeOpenThreadAdapter()

    thread = mock.Mock()
    thread.root = mock.Mock()
    thread.root.get_children.return_value = []

    thread_context = ChatThreadContext(
        session=AgentSessionContext(),
        thread=thread,
        path=".threads/main.thread",
    )

    first_participant = Participant(
        id="caller-id",
        attributes={"name": "alice", "current_file": "docs/plan.md"},
    )
    second_participant = Participant(
        id="caller-id",
        attributes={"name": "alice"},
    )

    await bot.on_chat_received(
        thread_context=thread_context,
        from_participant=first_participant,
        message={"text": "First"},
    )
    await bot.on_chat_received(
        thread_context=thread_context,
        from_participant=second_participant,
        message={"text": "Second"},
    )

    assert any(
        message.get("role") == "assistant"
        and message.get("content") == "alice is not currently viewing any files."
        for message in adapter.last_messages
    )


@pytest.mark.asyncio
async def test_on_chat_received_updates_thread_index_modified_at() -> None:
    adapter = _CaptureChatAdapter()
    bot = _ChatBotAlwaysReplies(llm_adapter=adapter, thread_dir="custom")
    room = _FakeRoom()
    bot._room = room
    bot._open_threads["custom/main.thread"] = _FakeOpenThreadAdapter()

    thread = mock.Mock()
    thread.root = mock.Mock()
    thread.root.get_children.return_value = []

    thread_context = ChatThreadContext(
        session=AgentSessionContext(),
        thread=thread,
        path="custom/main.thread",
    )
    from_participant = Participant(
        id="caller-id",
        attributes={"name": "alice"},
    )

    doc = _FakeThreadListDocument()
    doc.root.append_child(
        tag_name="thread",
        attributes={
            "name": "main",
            "path": "custom/main.thread",
            "created_at": "2024-01-01T00:00:00Z",
            "modified_at": "2024-01-01T00:00:00Z",
        },
    )
    bot._thread_list_document = doc

    result = await bot.on_chat_received(
        thread_context=thread_context,
        from_participant=from_participant,
        message={"text": "hello"},
    )

    assert result == "ok"
    entry = doc.root.get_children()[0]
    assert entry.get_attribute("created_at") == "2024-01-01T00:00:00Z"
    assert entry.get_attribute("modified_at") != "2024-01-01T00:00:00Z"


@pytest.mark.asyncio
async def test_should_reply_defaults_true_for_unstructured_decision_response() -> None:
    adapter = _ShouldReplyAdapter(response="Indexed reply")
    bot = ChatBot(llm_adapter=adapter)
    bot._room = _FakeRoom()

    thread = _FakeThreadDocument()
    members = thread.root.append_child(tag_name="members")
    members.append_child(tag_name="member", attributes={"name": "assistant"})
    members.append_child(tag_name="member", attributes={"name": "alice"})
    members.append_child(tag_name="member", attributes={"name": "bob"})

    context = ChatThreadContext(
        session=AgentSessionContext(),
        thread=thread,
        path=".threads/main.thread",
    )
    from_user = Participant(id="alice-id", attributes={"name": "alice"})
    online = [
        from_user,
        Participant(id="bob-id", attributes={"name": "bob"}),
    ]

    should_reply = await bot.should_reply(
        context=context,
        has_more_than_one_other_user=True,
        toolkits=[],
        from_user=from_user,
        online=online,
    )

    assert should_reply is True


@pytest.mark.asyncio
async def test_should_reply_parses_json_string_decision_response() -> None:
    adapter = _ShouldReplyAdapter(
        response='{"reasoning":"waiting on bob","expecting_assistant_reply":false,"next_user":"bob"}'
    )
    bot = ChatBot(llm_adapter=adapter)
    bot._room = _FakeRoom()

    thread = _FakeThreadDocument()
    members = thread.root.append_child(tag_name="members")
    members.append_child(tag_name="member", attributes={"name": "assistant"})
    members.append_child(tag_name="member", attributes={"name": "alice"})
    members.append_child(tag_name="member", attributes={"name": "bob"})

    context = ChatThreadContext(
        session=AgentSessionContext(),
        thread=thread,
        path=".threads/main.thread",
    )
    from_user = Participant(id="alice-id", attributes={"name": "alice"})
    online = [
        from_user,
        Participant(id="bob-id", attributes={"name": "bob"}),
    ]

    should_reply = await bot.should_reply(
        context=context,
        has_more_than_one_other_user=True,
        toolkits=[],
        from_user=from_user,
        online=online,
    )

    assert should_reply is False


def test_record_new_thread_in_index_adds_entry() -> None:
    adapter = _CaptureChatAdapter()
    bot = ChatBot(llm_adapter=adapter, thread_dir="custom")
    doc = _FakeThreadListDocument()
    bot._thread_list_document = doc

    path = "custom/release-plan.thread"
    bot._record_new_thread_in_index(path=path)

    children = doc.root.get_children()
    assert len(children) == 1
    entry = children[0]
    assert entry.tag_name == "thread"
    assert entry.get_attribute("name") == "Release Plan"
    assert entry.get_attribute("path") == path
    created_at = entry.get_attribute("created_at")
    modified_at = entry.get_attribute("modified_at")
    assert isinstance(created_at, str)
    assert isinstance(modified_at, str)
    assert created_at.endswith("Z")
    assert modified_at.endswith("Z")
    assert created_at == modified_at


def test_record_new_thread_in_index_uses_new_chat_for_guid_path() -> None:
    adapter = _CaptureChatAdapter()
    bot = ChatBot(llm_adapter=adapter, thread_dir="custom")
    doc = _FakeThreadListDocument()
    bot._thread_list_document = doc

    path = "custom/123e4567-e89b-12d3-a456-426614174000.thread"
    bot._record_new_thread_in_index(path=path)

    entry = doc.root.get_children()[0]
    assert entry.get_attribute("name") == "New Chat"


def test_touch_thread_in_index_updates_modified_at() -> None:
    adapter = _CaptureChatAdapter()
    bot = ChatBot(llm_adapter=adapter, thread_dir="custom")
    doc = _FakeThreadListDocument()
    path = "custom/main.thread"
    doc.root.append_child(
        tag_name="thread",
        attributes={
            "name": "main",
            "path": path,
            "created_at": "2024-01-01T00:00:00Z",
            "modified_at": "2024-01-01T00:00:00Z",
        },
    )
    bot._thread_list_document = doc

    bot._touch_thread_in_index(path=path)

    entry = doc.root.get_children()[0]
    assert entry.get_attribute("name") == "main"
    assert entry.get_attribute("created_at") == "2024-01-01T00:00:00Z"
    assert entry.get_attribute("modified_at") != "2024-01-01T00:00:00Z"


def test_touch_thread_in_index_forces_modified_at_forward_when_clock_does_not_advance() -> (
    None
):
    adapter = _CaptureChatAdapter()
    bot = ChatBot(llm_adapter=adapter, thread_dir="custom")
    doc = _FakeThreadListDocument()
    path = "custom/main.thread"
    existing_modified_at = "2024-01-01T00:00:00Z"
    doc.root.append_child(
        tag_name="thread",
        attributes={
            "name": "main",
            "path": path,
            "created_at": existing_modified_at,
            "modified_at": existing_modified_at,
        },
    )
    bot._thread_list_document = doc

    with mock.patch.object(bot, "_utc_now_iso", return_value=existing_modified_at):
        bot._touch_thread_in_index(path=path)

    entry = doc.root.get_children()[0]
    updated_modified_at = entry.get_attribute("modified_at")
    assert isinstance(updated_modified_at, str)
    assert bot._parse_iso_datetime(value=updated_modified_at) > bot._parse_iso_datetime(
        value=existing_modified_at
    )


def test_record_new_thread_in_index_is_noop_without_explicit_thread_dir() -> None:
    adapter = _CaptureChatAdapter()
    bot = ChatBot(llm_adapter=adapter)
    doc = _FakeThreadListDocument()
    bot._thread_list_document = doc

    bot._record_new_thread_in_index(path=".threads/assistant/test.thread")

    assert len(doc.root.get_children()) == 0


def test_chat_error_message_uses_room_exception_text() -> None:
    adapter = _CaptureChatAdapter()
    bot = ChatBot(llm_adapter=adapter)
    assert (
        bot._chat_error_message(error=RoomException("Your account is out of credits"))
        == "Your account is out of credits"
    )


def test_chat_error_message_uses_generic_for_non_room_exception() -> None:
    adapter = _CaptureChatAdapter()
    bot = ChatBot(llm_adapter=adapter)
    assert (
        bot._chat_error_message(error=RuntimeError("boom"))
        == "An unexpected error occured. Please try again later."
    )


def test_chat_error_message_uses_handshake_error_message() -> None:
    adapter = _CaptureChatAdapter()
    bot = ChatBot(llm_adapter=adapter)
    error = _make_ws_handshake_error(
        status=402,
        headers={
            "X-Meshagent-Error-Message": "Your account is out of credits. Add credits to continue.",
        },
    )
    assert (
        bot._chat_error_message(error=error)
        == "Your account is out of credits. Add credits to continue."
    )


@pytest.mark.asyncio
async def test_thread_list_document_open_close_uses_thread_dir_index_path() -> None:
    adapter = _CaptureChatAdapter()
    sync = _FakeSync()
    room = _FakeRoom(sync=sync)
    bot = ChatBot(llm_adapter=adapter, thread_dir="custom")
    bot._room = room

    await bot._open_thread_list_document()

    assert len(sync.open_calls) == 1
    assert sync.open_calls[0]["path"] == "custom/index.threadl"
    assert sync.open_calls[0]["schema"] is thread_list_schema
    assert bot._thread_list_document is sync.document

    await bot._close_thread_list_document(room=room)

    assert sync.close_calls == ["custom/index.threadl"]
    assert bot._thread_list_document is None
    assert bot._thread_list_path is None


@pytest.mark.asyncio
async def test_chatbot_start_sets_thread_attributes_on_participant() -> None:
    adapter = _CaptureChatAdapter()
    sync = _FakeSync()
    room = _FakeRoom(sync=sync)
    bot = ChatBot(
        llm_adapter=adapter,
        thread_dir="custom",
        threading_mode="default-new",
    )

    with mock.patch.object(
        bot,
        "get_exposed_toolkits",
        new=mock.AsyncMock(return_value=[]),
    ):
        await bot.start(room=room)

    assert (
        room.local_participant.get_attribute("meshagent.chatbot.thread-dir") == "custom"
    )
    assert (
        "meshagent.chatbot.thread-dir",
        "custom",
    ) in room.local_participant.set_attribute_calls
    assert (
        room.local_participant.get_attribute("meshagent.chatbot.thread-list")
        == "custom/index.threadl"
    )
    assert (
        "meshagent.chatbot.thread-list",
        "custom/index.threadl",
    ) in room.local_participant.set_attribute_calls

    await bot.stop()


@pytest.mark.asyncio
async def test_thread_list_document_not_opened_without_explicit_thread_dir() -> None:
    adapter = _CaptureChatAdapter()
    sync = _FakeSync()
    room = _FakeRoom(sync=sync)
    bot = ChatBot(llm_adapter=adapter)
    bot._room = room

    await bot._open_thread_list_document()

    assert sync.open_calls == []


@pytest.mark.asyncio
async def test_spawn_thread_waits_for_cleanup_before_returning() -> None:
    adapter = _CaptureChatAdapter()
    bot = ChatBot(llm_adapter=adapter)
    room = _FakeRoom()
    bot._room = room

    path = ".threads/assistant/main.thread"
    tracked_context = _TrackedThreadContext(path=path)
    bot._thread_contexts[path] = tracked_context  # type: ignore[assignment]

    close_started = asyncio.Event()
    allow_close = asyncio.Event()

    async def _close_thread(*, path: str) -> None:
        del path
        close_started.set()
        await allow_close.wait()

    bot.close_thread = mock.AsyncMock(side_effect=_close_thread)  # type: ignore[method-assign]
    bot._safe_invoke_thread_event = mock.AsyncMock()  # type: ignore[method-assign]

    task = asyncio.create_task(
        bot._spawn_thread(
            path=path,
            messages=_ExplodingMessageChannel(error=RuntimeError("thread ended")),  # type: ignore[arg-type]
        )
    )

    with pytest.raises(asyncio.TimeoutError):
        await asyncio.wait_for(asyncio.shield(task), timeout=0.05)

    assert close_started.is_set()
    assert tracked_context.closed is True
    assert path not in bot._thread_contexts

    allow_close.set()

    with pytest.raises(RuntimeError, match="thread ended"):
        await task


@pytest.mark.asyncio
async def test_stop_waits_for_thread_cleanup_before_returning() -> None:
    adapter = _CaptureChatAdapter()
    bot = ChatBot(llm_adapter=adapter)
    room = _FakeRoom()
    bot._room = room
    bot._exposed_toolkits = []

    path = ".threads/assistant/main.thread"
    close_started = asyncio.Event()
    allow_close = asyncio.Event()

    async def _close_thread(*, path: str) -> None:
        del path
        close_started.set()
        await allow_close.wait()

    bot.close_thread = mock.AsyncMock(side_effect=_close_thread)  # type: ignore[method-assign]
    bot._close_thread_list_document = mock.AsyncMock()  # type: ignore[method-assign]
    bot._clear_all_thread_statuses = mock.AsyncMock()  # type: ignore[method-assign]

    task = asyncio.create_task(
        bot._spawn_thread(
            path=path,
            messages=_BlockingMessageChannel(),  # type: ignore[arg-type]
        )
    )
    bot._thread_tasks[path] = task
    bot._message_channels[path] = mock.Mock(spec=Chan)

    stop_task = asyncio.create_task(bot.stop())

    with pytest.raises(asyncio.TimeoutError):
        await asyncio.wait_for(asyncio.shield(stop_task), timeout=0.05)

    assert close_started.is_set()

    allow_close.set()
    await stop_task
    assert task.done()



import re

import pytest
from typing import Optional
import meshagent.agents.llmrunner as llmrunner_module
from meshagent.agents.adapter import LLMAdapter
from meshagent.agents.context import AgentSessionContext
from meshagent.agents.llmrunner import LLMTaskRunner
from meshagent.agents.task_runner import TaskContext


class _FakeParticipant:
    def __init__(self, *, name: str, participant_id: str):
        self._name = name
        self.id = participant_id

    def get_attribute(self, key: str):
        if key == "name":
            return self._name
        return None


class _FakeElement:
    def __init__(self, *, tag_name: str, attributes: Optional[dict] = None):
        self.tag_name = tag_name
        self._attributes = dict(attributes or {})
        self._children: list["_FakeElement"] = []

    def get_attribute(self, key: str):
        return self._attributes.get(key)

    def set_attribute(self, key: str, value) -> None:
        self._attributes[key] = value

    def get_children(self):
        return self._children

    def append_child(self, tag_name: str, attributes: Optional[dict] = None):
        child = _FakeElement(tag_name=tag_name, attributes=attributes)
        self._children.append(child)
        return child


class _FakeThreadRoot:
    def __init__(self):
        self._members = _FakeElement(tag_name="members")

    def get_children_by_tag_name(self, tag_name: str):
        if tag_name == "members":
            return [self._members]
        return []

    @property
    def members(self) -> _FakeElement:
        return self._members


class _FakeThreadDocument:
    def __init__(self):
        self.root = _FakeThreadRoot()

    @property
    def member_names(self) -> list[str]:
        names: list[str] = []
        for child in self.root.members.get_children():
            if child.tag_name != "member":
                continue
            name = child.get_attribute("name")
            if isinstance(name, str):
                names.append(name)
        return names


class _FakeThreadListDocument:
    def __init__(self):
        self.root = _FakeElement(tag_name="thread_list")

    def get_state(self) -> bytes:
        return b"thread-list-state"


class _FakeSync:
    def __init__(self):
        self.document = _FakeThreadListDocument()
        self.open_calls: list[dict] = []
        self.close_calls: list[str] = []
        self.sync_calls: list[dict] = []

    async def open(self, *, path: str, schema=None):
        self.open_calls.append({"path": path, "schema": schema})
        return self.document

    async def close(self, *, path: str):
        self.close_calls.append(path)

    async def sync(self, *, path: str, data: bytes):
        self.sync_calls.append({"path": path, "data": data})


class _FakeStorage:
    def __init__(self, *, existing_paths: Optional[set[str]] = None):
        self._existing_paths = set(existing_paths or set())
        self.exists_calls: list[str] = []

    async def exists(self, *, path: str) -> bool:
        self.exists_calls.append(path)
        return path in self._existing_paths


class _FakeRoom:
    def __init__(self, *, existing_paths: Optional[set[str]] = None):
        self.local_participant = _FakeParticipant(
            name="assistant",
            participant_id="assistant-id",
        )
        self.sync = _FakeSync()
        self.storage = _FakeStorage(existing_paths=existing_paths)


class _FakeThreadAdapter:
    instances: list["_FakeThreadAdapter"] = []

    def __init__(self, *, room, path: str):
        del room
        self.path = path
        self.started = False
        self.stopped = False
        self.appended = False
        self.writes: list[tuple[str, str]] = []
        self.events: list[dict] = []
        self.thread = _FakeThreadDocument()
        _FakeThreadAdapter.instances.append(self)

    async def start(self) -> None:
        self.started = True

    async def stop(self) -> None:
        self.stopped = True

    def append_messages(self, *, context: AgentSessionContext) -> None:
        del context
        self.appended = True

    def write_text_message(self, *, text: str, participant) -> None:
        if isinstance(participant, str):
            participant_name = participant
        else:
            participant_name = participant.get_attribute("name") or ""

        self.writes.append((text, participant_name))

    def push(self, *, event: dict) -> None:
        self.events.append(event)


class _FakeLLMAdapter(LLMAdapter):
    def __init__(self, *, generated_thread_name: str = "release planning"):
        self.generated_thread_name = generated_thread_name
        self.calls: list[dict] = []

    def default_model(self) -> str:
        return "test-model"

    async def next(
        self,
        *,
        context,
        room,
        toolkits,
        output_schema=None,
        event_handler=None,
        steering_callback=None,
        model=None,
        on_behalf_of=None,
        options: Optional[dict] = None,
    ):
        del steering_callback
        self.calls.append(
            {
                "context": context,
                "room": room,
                "toolkits": toolkits,
                "output_schema": output_schema,
                "model": model,
                "on_behalf_of": on_behalf_of,
            }
        )

        if output_schema is not None:
            return {"thread_name": self.generated_thread_name}

        context.turn_count += 1

        if event_handler is not None:
            event_handler({"type": "response.content_part.added"})
            event_handler(
                {"type": "response.output_text.done", "text": "assistant response"}
            )

        return "assistant response"


def _make_context() -> TaskContext:
    room = _FakeRoom()
    caller = _FakeParticipant(name="caller", participant_id="caller-id")
    return TaskContext(
        session=AgentSessionContext(system_role=None),
        room=room,
        caller=caller,
        on_behalf_of=None,
        toolkits=[],
    )


def test_llm_task_runner_input_schema_path_only_for_manual_mode() -> None:
    adapter = _FakeLLMAdapter()
    manual_runner = LLMTaskRunner(llm_adapter=adapter, threading_mode="manual")
    auto_runner = LLMTaskRunner(llm_adapter=adapter, threading_mode="auto")
    none_runner = LLMTaskRunner(llm_adapter=adapter, threading_mode="none")

    assert "path" in manual_runner.input_schema["properties"]
    assert "path" not in auto_runner.input_schema["properties"]
    assert "path" not in none_runner.input_schema["properties"]


def test_llm_task_runner_fallback_thread_name_uses_timestamp() -> None:
    adapter = _FakeLLMAdapter()
    runner = LLMTaskRunner(llm_adapter=adapter, threading_mode="auto")
    thread_name = runner._fallback_thread_name(prompt="any prompt")

    assert re.fullmatch(r"thread-\d{8}-\d{6}", thread_name) is not None


@pytest.mark.asyncio
async def test_llm_task_runner_manual_threading_uses_input_path(monkeypatch) -> None:
    _FakeThreadAdapter.instances.clear()
    monkeypatch.setattr(llmrunner_module, "ThreadAdapter", _FakeThreadAdapter)

    adapter = _FakeLLMAdapter()
    runner = LLMTaskRunner(llm_adapter=adapter, threading_mode="manual")
    context = _make_context()

    async def _no_required_toolkits(*, context):
        del context
        return []

    monkeypatch.setattr(runner, "get_required_toolkits", _no_required_toolkits)

    result = await runner.ask(
        context=context,
        arguments={"prompt": "hello", "path": "/threads/manual"},
    )

    assert result == "assistant response"
    assert len(adapter.calls) == 1

    assert len(_FakeThreadAdapter.instances) == 1
    thread_adapter = _FakeThreadAdapter.instances[0]
    assert thread_adapter.path == "/threads/manual"
    assert thread_adapter.started
    assert thread_adapter.appended
    assert thread_adapter.stopped
    assert thread_adapter.thread.member_names == ["assistant"]
    assert thread_adapter.writes == [("hello", "caller")]
    assert [event["type"] for event in thread_adapter.events] == [
        "response.content_part.added",
        "response.output_text.done",
    ]
    assert context.room.sync.open_calls == []
    assert context.room.sync.close_calls == []


@pytest.mark.asyncio
async def test_llm_task_runner_auto_threading_generates_path(monkeypatch) -> None:
    _FakeThreadAdapter.instances.clear()
    monkeypatch.setattr(llmrunner_module, "ThreadAdapter", _FakeThreadAdapter)

    adapter = _FakeLLMAdapter(generated_thread_name="Release Planning / Q1")
    runner = LLMTaskRunner(
        llm_adapter=adapter,
        threading_mode="auto",
        thread_dir="/threads",
    )
    context = _make_context()

    async def _no_required_toolkits(*, context):
        del context
        return []

    monkeypatch.setattr(runner, "get_required_toolkits", _no_required_toolkits)

    result = await runner.ask(
        context=context,
        arguments={"prompt": "Plan the Q1 release milestones"},
    )

    assert result == "assistant response"
    assert len(adapter.calls) == 2

    thread_name_call = adapter.calls[0]
    assert thread_name_call["output_schema"] is not None
    assert thread_name_call["output_schema"]["required"] == ["thread_name"]

    assert len(_FakeThreadAdapter.instances) == 1
    thread_adapter = _FakeThreadAdapter.instances[0]
    assert thread_adapter.path == "/threads/release-planning-q1.thread"
    assert thread_adapter.thread.member_names == ["assistant"]
    assert thread_adapter.writes == [("Plan the Q1 release milestones", "caller")]
    assert context.room.sync.open_calls[0]["path"] == "/threads/index.threadl"
    assert context.room.sync.sync_calls[0]["path"] == "/threads/index.threadl"
    assert context.room.sync.sync_calls[0]["data"] == b"dGhyZWFkLWxpc3Qtc3RhdGU="
    assert context.room.sync.close_calls == ["/threads/index.threadl"]
    thread_entries = context.room.sync.document.root.get_children()
    assert len(thread_entries) == 1
    assert thread_entries[0].tag_name == "thread"
    assert (
        thread_entries[0].get_attribute("path") == "/threads/release-planning-q1.thread"
    )


@pytest.mark.asyncio
async def test_llm_task_runner_auto_threading_uses_custom_name_rules(
    monkeypatch,
) -> None:
    _FakeThreadAdapter.instances.clear()
    monkeypatch.setattr(llmrunner_module, "ThreadAdapter", _FakeThreadAdapter)

    adapter = _FakeLLMAdapter(generated_thread_name="custom thread")
    runner = LLMTaskRunner(
        llm_adapter=adapter,
        threading_mode="auto",
        thread_dir="/threads",
        thread_name_rules=["pick a kebab-case name from this task prompt"],
    )
    context = _make_context()

    async def _no_required_toolkits(*, context):
        del context
        return []

    monkeypatch.setattr(runner, "get_required_toolkits", _no_required_toolkits)

    result = await runner.ask(
        context=context,
        arguments={"prompt": "Do custom naming"},
    )

    assert result == "assistant response"
    assert len(adapter.calls) == 2
    thread_name_call = adapter.calls[0]
    assert thread_name_call["context"].instructions is not None
    assert (
        "pick a kebab-case name from this task prompt"
        in thread_name_call["context"].instructions
    )



from __future__ import annotations

import asyncio
import contextlib
import logging
import mimetypes
import uuid
from dataclasses import dataclass, field
from datetime import datetime, timezone
from pathlib import PurePosixPath
from typing import Any, Awaitable, Callable, Literal, Optional, TypeVar
from urllib.parse import urlparse

from meshagent.api import Participant, RoomClient
from meshagent.api.messaging import FileContent
from meshagent.agents.adapter import LLMAdapter, ToolCallApprovalRequest
from meshagent.agents.context import AgentSessionContext
from meshagent.tools import RemoteToolkit, ToolContext, Toolkit, ToolkitBuilder
from .process_thread_adapter import AgentProcessThreadAdapter
from .thread_adapter import ThreadAdapter, default_format_message
from .messages import (
    AGENT_EVENT_THREAD_CLEARED,
    AGENT_EVENT_TOOL_CALL_APPROVAL_REQUESTED,
    AGENT_EVENT_TURN_ENDED,
    AGENT_EVENT_TURN_INTERRUPTED,
    AGENT_EVENT_TURN_INTERRUPT_ACCEPTED,
    AGENT_EVENT_TURN_START_ACCEPTED,
    AGENT_EVENT_TURN_STARTED,
    AGENT_EVENT_TURN_STEER_ACCEPTED,
    AGENT_EVENT_TURN_STEERED,
    AGENT_EVENT_TURN_STEER_REJECTED,
    AGENT_MESSAGE_THREAD_CLEAR,
    AGENT_MESSAGE_TOOL_CALL_APPROVE,
    AGENT_MESSAGE_TOOL_CALL_REJECT,
    AGENT_MESSAGE_TURN_INTERRUPT,
    AGENT_MESSAGE_TURN_START,
    AGENT_MESSAGE_TURN_STEER,
    AgentError,
    AgentFileContent,
    AgentMessage,
    AgentTextContent,
    AgentToolCallApprovalRequested,
    ApproveAgentToolCall,
    ClearThread,
    RejectAgentToolCall,
    ThreadCleared,
    TurnEnded,
    TurnInterrupt,
    TurnInterrupted,
    TurnInterruptAccepted,
    TurnStartAccepted,
    TurnStart,
    TurnSteerAccepted,
    TurnSteered,
    TurnSteer,
    TurnSteerRejected,
    TurnStarted,
)

logger = logging.getLogger("agent-process")
_THREAD_STATUS_ACTIVE_STATES = {
    "queued",
    "in_progress",
    "running",
    "pending",
    "searching",
}
_THREAD_STATUS_TERMINAL_STATES = {"completed", "failed", "cancelled"}
_THREAD_ADAPTER_REQUEST_MESSAGE_TYPES = frozenset(
    {
        AGENT_MESSAGE_THREAD_CLEAR,
        AGENT_MESSAGE_TOOL_CALL_APPROVE,
        AGENT_MESSAGE_TOOL_CALL_REJECT,
        AGENT_MESSAGE_TURN_INTERRUPT,
        AGENT_MESSAGE_TURN_START,
        AGENT_MESSAGE_TURN_STEER,
    }
)

LifecycleState = Literal["stopped", "starting", "started", "stopping", "failed"]
ChannelState = LifecycleState
SessionInitializer = Callable[[], Awaitable[AgentSessionContext]]
TurnInstructionsProvider = Callable[[Participant | None], Awaitable[str | None]]
TurnToolkitsBuilder = Callable[
    [Participant | None, str, list["TurnStart | TurnSteer"]],
    Awaitable[list[Toolkit]],
]


@dataclass(slots=True)
class Message:
    data: AgentMessage
    sender: Participant | None = None
    source: Channel | AgentProcess | None = None


_MessageT = TypeVar("_MessageT", bound=AgentMessage)


def _coerce_message_data(data: AgentMessage, model: type[_MessageT]) -> _MessageT:
    if isinstance(data, model):
        return data

    return model.model_validate(data.model_dump(mode="python"))


class Channel:
    def __init__(self) -> None:
        self._supervisor: AgentSupervisor | None = None
        self._state: ChannelState = "stopped"
        self._run_task: asyncio.Task[None] | None = None
        self._stop = asyncio.Event()
        self._queue: asyncio.Queue[Message] = asyncio.Queue()
        self._lifecycle_lock = asyncio.Lock()

    @property
    def state(self) -> ChannelState:
        return self._state

    @property
    def supervisor(self) -> AgentSupervisor | None:
        return self._supervisor

    def handles(self, message: Message) -> bool:
        del message
        return True

    def send(self, message: Message) -> None:
        if self._state != "started" or self._supervisor is None:
            logger.debug("dropping channel message while channel is not started")
            return

        if self._stop.is_set():
            logger.debug("dropping channel message during shutdown")
            return

        if self.handles(message):
            try:
                self._queue.put_nowait(message)
            except asyncio.QueueShutDown:
                logger.debug("dropping channel message after queue shutdown")

    def emit(self, *, sender: Participant | None, payload: AgentMessage) -> None:
        supervisor = self.supervisor
        if supervisor is None:
            return

        supervisor.send(Message(data=payload, sender=sender, source=self))

    def get_agent_toolkits(self) -> list[Toolkit]:
        return []

    def get_exposed_toolkits(self) -> list[RemoteToolkit]:
        return []

    async def on_start(self) -> None:
        return None

    async def on_message(self, message: Message) -> None:
        del message
        return None

    async def on_stop(self) -> None:
        return None

    async def run(self) -> None:
        while not self._stop.is_set():
            with contextlib.suppress(asyncio.QueueShutDown):
                message = await self._queue.get()
                await self.on_message(message)

    async def start(self, supervisor: AgentSupervisor) -> None:
        async with self._lifecycle_lock:
            if self._state not in {"stopped", "failed"}:
                raise ValueError("already started")

            if self._supervisor is not None:
                raise ValueError("already started")

            self._state = "starting"
            self._stop.clear()

            try:
                self._supervisor = supervisor

                await self.on_start()

                self._run_task = asyncio.create_task(self.run())
                self._state = "started"
            except Exception:
                self._state = "failed"
                self._supervisor = None
                self._run_task = None
                raise

    async def stop(self, supervisor: AgentSupervisor) -> None:
        async with self._lifecycle_lock:
            if self._supervisor is None or self._supervisor is not supervisor:
                raise ValueError("not started")

            self._state = "stopping"

            try:
                self._stop.set()
                self._queue.shutdown()
                if self._run_task is not None:
                    await self._run_task
                await self.on_stop()
            except Exception:
                logger.exception("channel failed during stop")
                self._state = "failed"
                self._supervisor = None
                self._run_task = None
                raise
            else:
                self._supervisor = None
                self._run_task = None
                self._state = "stopped"


SupervisorState = LifecycleState


class AgentSupervisor:
    def __init__(self) -> None:
        self.channels: list[Channel] = []
        self.processes: list[AgentProcess] = []
        self._stop = asyncio.Event()
        self._state: SupervisorState = "stopped"
        self._run_task: asyncio.Task[None] | None = None
        self._queue: asyncio.Queue[Message] = asyncio.Queue()
        self._lifecycle_lock = asyncio.Lock()

    @property
    def state(self) -> SupervisorState:
        return self._state

    def add_channel(self, channel: Channel) -> None:
        self.channels.append(channel)

    def stop_channel(self, channel: Channel) -> None:
        if channel in self.channels:
            self.channels.remove(channel)

    def add_process(self, process: AgentProcess) -> None:
        self.processes.append(process)

    def create_thread_process(self, thread_id: str) -> AgentProcess:
        raise NotImplementedError(
            f"{self.__class__.__name__} must implement create_thread_process"
        )

    def send(self, message: Message) -> None:
        if self._stop.is_set() or self._state == "stopping":
            logger.debug("dropping supervisor message during shutdown")
            return

        try:
            self._queue.put_nowait(message)
        except asyncio.QueueShutDown:
            logger.debug("dropping supervisor message after queue shutdown")

    def emit(
        self,
        *,
        sender: Participant | None,
        payload: AgentMessage,
    ) -> None:
        self.send(Message(data=payload, sender=sender))

    async def on_start(self) -> None:
        return None

    async def on_stop(self) -> None:
        return None

    async def start(self) -> None:
        async with self._lifecycle_lock:
            if self._state not in {"stopped", "failed"}:
                raise ValueError("already started")

            self._state = "starting"
            self._stop.clear()

            try:
                await self.on_start()
                await self._ensure_children_started()
                self._run_task = asyncio.create_task(self.run())
                self._state = "started"
            except Exception:
                logger.exception("agent supervisor failed during start")
                self._state = "failed"
                self._run_task = None
                raise

    async def stop(self) -> None:
        async with self._lifecycle_lock:
            if self._run_task is None:
                raise ValueError("not started")

            self._state = "stopping"

            try:
                self._stop.set()
                self._queue.shutdown()
                await self._run_task
                await self.on_stop()
            except Exception:
                logger.exception("agent supervisor failed during stop")
                self._state = "failed"
                self._run_task = None
                raise
            else:
                self._run_task = None
                self._state = "stopped"

    def _process_for_thread(self, *, thread_id: str) -> AgentProcess | None:
        for process in self.processes:
            if process.thread_id == thread_id:
                return process
        return None

    def _process_for_turn(self, *, turn_id: str) -> AgentProcess | None:
        for process in self.processes:
            if process.turn_id == turn_id:
                return process
        return None

    @staticmethod
    def _copy_message(message: Message, *, data: AgentMessage) -> Message:
        return Message(
            data=data,
            sender=message.sender,
            source=message.source,
        )

    def _send_to_channels(self, message: Message) -> None:
        for channel in self.channels:
            if message.source is channel:
                continue
            channel.send(message)

    def _send_to_processes(
        self,
        message: Message,
        *,
        processes: list[AgentProcess] | None = None,
    ) -> None:
        target_processes = self.processes if processes is None else processes
        for process in target_processes:
            if message.source is process:
                continue
            process.send(message)

    async def _ensure_routing_processes_started(
        self,
        *,
        processes: list[AgentProcess] | None,
    ) -> list[AgentProcess] | None:
        if processes is None:
            return None

        started_processes: list[AgentProcess] = []
        for process in processes:
            if process.state == "stopped":
                try:
                    await process.start(self)
                except Exception:
                    logger.exception(
                        "process %s failed during routed start; dropping message",
                        process.__class__.__name__,
                    )
                    continue

            if process.state == "started" and process.supervisor is self:
                started_processes.append(process)

        return started_processes

    async def _route(self, message: Message) -> None:
        routed_message = message
        target_processes: list[AgentProcess] | None = None
        message_type = message.data.type
        if message_type == AGENT_MESSAGE_TURN_START:
            turn_start = _coerce_message_data(message.data, TurnStart)
            process = self._process_for_thread(thread_id=turn_start.thread_id)
            if process is None:
                process = self.create_thread_process(turn_start.thread_id)
                self.add_process(process)

            routed_message = self._copy_message(
                message,
                data=turn_start,
            )
            target_processes = [process]

        elif message_type == AGENT_MESSAGE_TURN_STEER:
            turn_steer = _coerce_message_data(message.data, TurnSteer)
            process = self._process_for_thread(thread_id=turn_steer.thread_id)
            routed_message = self._copy_message(
                message,
                data=turn_steer,
            )
            if process is not None and process.turn_id == turn_steer.turn_id:
                target_processes = [process]
            else:
                target_processes = []

        elif message_type == AGENT_MESSAGE_TURN_INTERRUPT:
            turn_interrupt = _coerce_message_data(message.data, TurnInterrupt)
            process = self._process_for_thread(thread_id=turn_interrupt.thread_id)
            routed_message = self._copy_message(
                message,
                data=turn_interrupt,
            )
            if process is not None and process.turn_id == turn_interrupt.turn_id:
                target_processes = [process]
            else:
                target_processes = []

        elif message_type == AGENT_MESSAGE_TOOL_CALL_APPROVE:
            approval = _coerce_message_data(message.data, ApproveAgentToolCall)
            process = self._process_for_thread(thread_id=approval.thread_id)
            routed_message = self._copy_message(
                message,
                data=approval,
            )
            if process is not None and process.turn_id == approval.turn_id:
                target_processes = [process]
            else:
                target_processes = []

        elif message_type == AGENT_MESSAGE_TOOL_CALL_REJECT:
            rejection = _coerce_message_data(message.data, RejectAgentToolCall)
            process = self._process_for_thread(thread_id=rejection.thread_id)
            routed_message = self._copy_message(
                message,
                data=rejection,
            )
            if process is not None and process.turn_id == rejection.turn_id:
                target_processes = [process]
            else:
                target_processes = []

        elif message_type == AGENT_MESSAGE_THREAD_CLEAR:
            clear_thread = _coerce_message_data(message.data, ClearThread)
            process = self._process_for_thread(thread_id=clear_thread.thread_id)
            if process is None:
                process = self.create_thread_process(clear_thread.thread_id)
                self.add_process(process)

            routed_message = self._copy_message(
                message,
                data=clear_thread,
            )
            target_processes = [process]

        target_processes = await self._ensure_routing_processes_started(
            processes=target_processes
        )
        self._send_to_channels(routed_message)
        self._send_to_processes(routed_message, processes=target_processes)

    async def _ensure_children_started(self) -> None:
        for channel in self.channels:
            if channel.state == "stopped":
                try:
                    await channel.start(self)
                except Exception:
                    logger.exception(
                        "channel %s failed during start; continuing",
                        channel.__class__.__name__,
                    )

        for process in self.processes:
            if process.state == "stopped":
                try:
                    await process.start(self)
                except Exception:
                    logger.exception(
                        "process %s failed during start; continuing",
                        process.__class__.__name__,
                    )

    async def _stop_children(self) -> None:
        errors: list[Exception] = []

        for process in self.processes:
            if process.supervisor is not self:
                continue
            if process.state != "stopped":
                try:
                    await process.stop(self)
                except (
                    Exception
                ) as exc:  # pragma: no cover - error path tested via caller state
                    errors.append(exc)

        for channel in self.channels:
            if channel.supervisor is not self:
                continue
            if channel.state != "stopped":
                try:
                    await channel.stop(self)
                except (
                    Exception
                ) as exc:  # pragma: no cover - error path tested via caller state
                    errors.append(exc)

        if errors:
            raise errors[0]

    async def run(self) -> None:
        try:
            while not self._stop.is_set():
                await self._ensure_children_started()

                with contextlib.suppress(asyncio.QueueShutDown):
                    message = await self._queue.get()
                    await self._route(message)
        finally:
            await self._stop_children()


ProcessState = LifecycleState


class AgentProcess:
    def __init__(
        self,
        supervisor: AgentSupervisor | None = None,
        *,
        thread_id: str | None = None,
        thread_adapter: ThreadAdapter | None = None,
    ) -> None:
        del supervisor
        if thread_adapter is not None:
            if thread_id is None:
                thread_id = thread_adapter.path
            elif thread_adapter.path != thread_id:
                raise ValueError("thread_adapter path must match thread_id")

        self._supervisor: AgentSupervisor | None = None
        self._thread_id = thread_id
        self._thread_adapter = thread_adapter
        self._state: ProcessState = "stopped"
        self._run_task: asyncio.Task[None] | None = None
        self._stop = asyncio.Event()
        self._queue: asyncio.Queue[Message] = asyncio.Queue()
        self._lifecycle_lock = asyncio.Lock()

    def send(self, message: Message) -> None:
        if self._state != "started" or self._supervisor is None:
            logger.debug("dropping process message while process is not started")
            return

        if self._stop.is_set():
            logger.debug("dropping process message during shutdown")
            return

        if self._should_mirror_to_thread_adapter(message=message):
            thread_adapter = self._thread_adapter
            if thread_adapter is not None:
                thread_adapter.push_message(
                    message=message.data,
                    sender=message.sender,
                )

        if self.handles(message):
            try:
                self._queue.put_nowait(message)
            except asyncio.QueueShutDown:
                logger.debug("dropping process message after queue shutdown")

    def handles(self, message: Message) -> bool:
        return False

    def _should_mirror_to_thread_adapter(self, *, message: Message) -> bool:
        thread_adapter = self._thread_adapter
        if thread_adapter is None:
            return False

        thread_id = self._thread_id
        if thread_id is None or message.data.thread_id != thread_id:
            return False

        if message.source is self:
            return False

        message_type = message.data.type
        if not message_type.startswith("meshagent.agent."):
            return False

        return message_type not in _THREAD_ADAPTER_REQUEST_MESSAGE_TYPES

    @property
    def state(self) -> ProcessState:
        return self._state

    @property
    def supervisor(self) -> AgentSupervisor | None:
        return self._supervisor

    @property
    def thread_id(self) -> str | None:
        return self._thread_id

    @property
    def thread_adapter(self) -> ThreadAdapter | None:
        return self._thread_adapter

    @property
    def turn_id(self) -> str | None:
        return None

    def emit(self, *, sender: Participant | None, payload: AgentMessage) -> None:
        supervisor = self.supervisor
        if supervisor is None:
            return

        supervisor.send(Message(data=payload, sender=sender, source=self))

    async def on_start(self) -> None:
        return None

    async def on_stop(self) -> None:
        return None

    async def start(self, supervisor: AgentSupervisor) -> None:
        async with self._lifecycle_lock:
            if self._state not in {"stopped", "failed"}:
                raise ValueError("already started")

            if self._supervisor is not None:
                raise ValueError("already started")

            self._state = "starting"
            self._stop.clear()

            try:
                self._supervisor = supervisor
                if self._thread_adapter is not None:
                    await self._thread_adapter.start()
                await self.on_start()
                self._run_task = asyncio.create_task(self.run())
                self._state = "started"
            except Exception:
                self._state = "failed"
                if self._thread_adapter is not None:
                    with contextlib.suppress(Exception):
                        await self._thread_adapter.stop()
                self._supervisor = None
                self._run_task = None
                raise

    async def on_message(self, message: Message) -> None:
        del message
        return None

    async def run(self) -> None:
        while not self._stop.is_set():
            with contextlib.suppress(asyncio.QueueShutDown):
                message = await self._queue.get()
                await self.on_message(message)

    async def stop(self, supervisor: AgentSupervisor) -> None:
        async with self._lifecycle_lock:
            if self._supervisor is None or self._supervisor is not supervisor:
                raise ValueError("not started")

            self._state = "stopping"

            try:
                self._stop.set()
                self._queue.shutdown()
                if self._run_task is not None:
                    await self._run_task
                await self.on_stop()
                if self._thread_adapter is not None:
                    await self._thread_adapter.stop()
            except Exception:
                logger.exception("agent process failed during stop")
                self._state = "failed"
                self._supervisor = None
                self._run_task = None
                raise
            else:
                self._supervisor = None
                self._run_task = None
                self._state = "stopped"


@dataclass(slots=True)
class _QueuedTurn:
    sender: Participant | None
    request: TurnStart
    queued_messages: list[_QueuedTurnMessage] = field(default_factory=list)


@dataclass(slots=True)
class _QueuedTurnMessage:
    sender: Participant | None
    request: TurnStart | TurnSteer


class LLMAgentProcess(AgentProcess):
    def __init__(
        self,
        *,
        thread_id: str,
        room: RoomClient,
        llm_adapter: LLMAdapter,
        toolkit_builders: Optional[list[ToolkitBuilder]] = None,
        toolkits: Optional[list[Toolkit]] = None,
        thread_adapter: AgentProcessThreadAdapter | None = None,
        session_initializer: SessionInitializer | None = None,
        turn_instructions_provider: TurnInstructionsProvider | None = None,
        turn_toolkits_builder: TurnToolkitsBuilder | None = None,
    ) -> None:
        if thread_adapter is not None and not isinstance(
            thread_adapter, AgentProcessThreadAdapter
        ):
            raise TypeError("thread_adapter must be an AgentProcessThreadAdapter")

        super().__init__(thread_id=thread_id, thread_adapter=thread_adapter)
        self.llm_adapter = llm_adapter
        self._turn_id: str | None = None
        self._handlers: dict[str, Callable[[Message], Awaitable[None]]] = {
            AGENT_MESSAGE_TURN_START: self.on_turn_start,
            AGENT_MESSAGE_TURN_STEER: self.on_turn_steer,
            AGENT_MESSAGE_TURN_INTERRUPT: self.on_turn_interrupt,
            AGENT_MESSAGE_THREAD_CLEAR: self.on_clear_thread,
            AGENT_MESSAGE_TOOL_CALL_APPROVE: self.on_tool_call_approve,
            AGENT_MESSAGE_TOOL_CALL_REJECT: self.on_tool_call_reject,
        }
        self._session_context: AgentSessionContext | None = None
        self._turn_task: asyncio.Task[None] | None = None
        self._active_next_task: asyncio.Task[Any] | None = None
        self._pending_turns: asyncio.Queue[_QueuedTurn] = asyncio.Queue()
        self._priority_turn: _QueuedTurn | None = None
        self._active_turn_queue: asyncio.Queue[_QueuedTurnMessage] | None = None
        self._toolkit_builders = list(toolkit_builders or [])
        self._toolkits = list(toolkits or [])
        self._room = room
        self._pending_tool_call_approvals: dict[str, asyncio.Future[bool]] = {}
        self._active_turn_sender: Participant | None = None
        self._pending_status_messages: list[_QueuedTurnMessage] = []
        self._interrupt_requested_turn_id: str | None = None
        self._interrupt_source_message_id: str | None = None
        self._session_initializer = session_initializer
        self._turn_instructions_provider = turn_instructions_provider
        self._turn_toolkits_builder = turn_toolkits_builder
        self.llm_adapter.set_tool_call_approval_handler(
            self._request_tool_call_approval
        )

    @property
    def turn_id(self) -> str | None:
        return self._turn_id

    @property
    def session_context(self) -> AgentSessionContext | None:
        return self._session_context

    @property
    def thread_adapter(self) -> AgentProcessThreadAdapter | None:
        adapter = super().thread_adapter
        if adapter is None:
            return None

        if not isinstance(adapter, AgentProcessThreadAdapter):
            raise TypeError("thread_adapter must be an AgentProcessThreadAdapter")

        return adapter

    @property
    def toolkits(self) -> list[Toolkit]:
        return self._toolkits

    @property
    def toolkit_builders(self) -> list[ToolkitBuilder]:
        return self._toolkit_builders

    @property
    def room(self) -> RoomClient:
        return self._room

    def emit(self, *, sender: Participant | None, payload: AgentMessage) -> None:
        thread_adapter = self.thread_adapter
        if thread_adapter is not None:
            thread_adapter.push_message(message=payload, sender=sender)

        super().emit(sender=sender, payload=payload)

    def handles(self, message: Message) -> bool:
        message_type = message.data.type
        if message_type not in self._handlers:
            return False

        return message.data.thread_id == self._thread_id

    async def on_session_context_created(self) -> None:
        if self._session_initializer is None:
            return None

        session_context = self.session_context
        if session_context is None:
            return None

        initialized_context = await self._session_initializer()
        session_context.messages.extend(initialized_context.messages)
        session_context.previous_messages.extend(initialized_context.previous_messages)
        session_context.previous_response_id = initialized_context.previous_response_id
        session_context.instructions = initialized_context.instructions
        return None

    # used to restore any persisted agent state
    async def on_restore_session_context(
        self,
        turn_id: str,
        session_context: AgentSessionContext,
    ) -> None:
        del turn_id
        del session_context
        return None

    async def ensure_session_context(self, *, turn_id: str) -> AgentSessionContext:
        if self._session_context is None:
            self._session_context = self.llm_adapter.create_session()
            await self.on_session_context_created()
            thread_adapter = self.thread_adapter
            if thread_adapter is not None:
                thread_adapter.restore_session_context(context=self._session_context)
            await self.on_restore_session_context(turn_id, self._session_context)
            await self._session_context.start()

        return self._session_context

    def _record_applied_turns(
        self,
        *,
        queued_messages: list[_QueuedTurnMessage],
    ) -> None:
        thread_adapter = self.thread_adapter
        if thread_adapter is None:
            return

        for queued_message in queued_messages:
            thread_adapter.push_message(
                message=queued_message.request,
                sender=queued_message.sender,
            )

    @staticmethod
    def _sender_name(sender: Participant | None) -> str | None:
        if sender is None:
            return None

        raw_name = sender.get_attribute("name")
        if not isinstance(raw_name, str):
            return None

        name = raw_name.strip()
        if name == "":
            return None

        return name

    def _format_live_turn_message(
        self,
        *,
        sender: Participant | None,
        message: str,
    ) -> str:
        sender_name = self._sender_name(sender)
        if sender_name is None or message == "":
            return message

        iso_timestamp = datetime.now(timezone.utc).isoformat().replace("+00:00", "Z")
        thread_adapter = self.thread_adapter
        if thread_adapter is not None:
            return thread_adapter.format_message(
                user_name=sender_name,
                message=message,
                iso_timestamp=iso_timestamp,
            )

        return default_format_message(
            user_name=sender_name,
            message=message,
            iso_timestamp=iso_timestamp,
        )

    def _file_attachment_message(
        self,
        *,
        sender: Participant | None,
        url: str,
    ) -> str:
        sender_name = self._sender_name(sender)
        if sender_name is None:
            return f"the user attached a file available at {url}"

        return f"{sender_name} attached a file available at {url}"

    @classmethod
    def _pending_status_message_payload(
        cls,
        *,
        queued_message: _QueuedTurnMessage,
    ) -> dict[str, Any]:
        request = queued_message.request
        content = request.model_dump(mode="json").get("content", [])
        return {
            "message_id": request.message_id,
            "message_type": request.type,
            "sender_name": cls._sender_name(queued_message.sender),
            "content": content if isinstance(content, list) else [],
        }

    async def _sync_pending_status_messages(self) -> None:
        thread_adapter = self.thread_adapter
        if thread_adapter is None:
            return

        await thread_adapter.set_pending_messages(
            pending_messages=[
                self._pending_status_message_payload(queued_message=queued_message)
                for queued_message in self._pending_status_messages
            ]
        )

    async def _add_pending_status_messages(
        self,
        *,
        queued_messages: list[_QueuedTurnMessage],
    ) -> None:
        if len(queued_messages) == 0:
            return

        self._pending_status_messages.extend(queued_messages)
        await self._sync_pending_status_messages()

    async def _remove_pending_status_messages(
        self,
        *,
        queued_messages: list[_QueuedTurnMessage],
    ) -> None:
        if len(queued_messages) == 0 or len(self._pending_status_messages) == 0:
            return

        removed_message_ids = {
            queued_message.request.message_id for queued_message in queued_messages
        }
        remaining_messages = [
            queued_message
            for queued_message in self._pending_status_messages
            if queued_message.request.message_id not in removed_message_ids
        ]
        if len(remaining_messages) == len(self._pending_status_messages):
            return

        self._pending_status_messages = remaining_messages
        await self._sync_pending_status_messages()

    async def _clear_pending_status_messages(self) -> None:
        if len(self._pending_status_messages) == 0:
            return

        self._pending_status_messages.clear()
        await self._sync_pending_status_messages()

    def _cancel_pending_tool_call_approvals(self) -> None:
        pending_approvals = list(self._pending_tool_call_approvals.values())
        self._pending_tool_call_approvals.clear()
        for future in pending_approvals:
            if not future.done():
                future.cancel()

    async def _request_tool_call_approval(
        self,
        context: ToolContext,
        request: ToolCallApprovalRequest,
    ) -> bool:
        del context

        turn_id = self._turn_id
        thread_id = self.thread_id
        if turn_id is None or thread_id is None:
            raise RuntimeError("tool call approval requested without an active turn")

        approval_future: asyncio.Future[bool] = (
            asyncio.get_running_loop().create_future()
        )
        self._pending_tool_call_approvals[request.item_id] = approval_future
        self.emit(
            sender=self._active_turn_sender,
            payload=AgentToolCallApprovalRequested(
                type=AGENT_EVENT_TOOL_CALL_APPROVAL_REQUESTED,
                thread_id=thread_id,
                turn_id=turn_id,
                item_id=request.item_id,
                toolkit=request.toolkit,
                tool=request.tool,
                arguments=request.arguments,
            ),
        )

        try:
            return await approval_future
        finally:
            existing_future = self._pending_tool_call_approvals.get(request.item_id)
            if existing_future is approval_future:
                del self._pending_tool_call_approvals[request.item_id]

    async def _resolve_tool_call_approval(
        self, *, item_id: str, approved: bool
    ) -> None:
        approval_future = self._pending_tool_call_approvals.get(item_id)
        if approval_future is None or approval_future.done():
            return

        approval_future.set_result(approved)

    async def _resolve_requested_toolkits(
        self,
        *,
        model: str,
        requested_toolkits: list[dict[str, Any]],
    ) -> list[Toolkit]:
        toolkits: list[Toolkit] = []
        for raw_config in requested_toolkits:
            toolkit_name = raw_config.get("name")
            if not isinstance(toolkit_name, str):
                raise ValueError("toolkit config must include a string `name`")

            matching_builder: ToolkitBuilder | None = None
            for builder in self._toolkit_builders:
                if builder.name == toolkit_name:
                    matching_builder = builder
                    break

            if matching_builder is None:
                raise ValueError(f"tool cannot be configured: {raw_config}")

            typed_config = matching_builder.type.model_validate(raw_config)
            toolkits.append(
                await matching_builder.make(
                    room=self._room,
                    model=model,
                    config=typed_config,
                )
            )

        return toolkits

    async def _build_turn_toolkits(
        self,
        *,
        model: str,
        turns: list[TurnStart | TurnSteer],
        sender: Participant | None = None,
    ) -> list[Toolkit]:
        if self._turn_toolkits_builder is not None:
            return await self._turn_toolkits_builder(sender, model, turns)

        combined_toolkits = [*self._toolkits]
        supervisor = self.supervisor
        if supervisor is not None:
            for channel in supervisor.channels:
                combined_toolkits.extend(channel.get_agent_toolkits())
        for turn in turns:
            if turn.toolkits is not None and len(turn.toolkits) > 0:
                combined_toolkits.extend(
                    await self._resolve_requested_toolkits(
                        model=model,
                        requested_toolkits=turn.toolkits,
                    )
                )
        return combined_toolkits

    @staticmethod
    def _normalize_room_storage_path(*, url: str) -> str:
        parsed_url = urlparse(url)
        if parsed_url.scheme != "room":
            raise ValueError(f"unsupported room file url: {url}")

        raw_path = f"{parsed_url.netloc}{parsed_url.path}".lstrip("/")
        normalized = PurePosixPath("/" + raw_path).as_posix().strip("/")
        if normalized == "":
            raise ValueError("room file url must reference a non-root storage path")

        if any(part in {".", ".."} for part in PurePosixPath(normalized).parts):
            raise ValueError("room file url cannot contain '.' or '..' segments")

        return normalized

    @staticmethod
    def _guess_url_mime_type(*, url: str) -> str | None:
        guessed_mime_type, _ = mimetypes.guess_type(urlparse(url).path)
        return guessed_mime_type

    def _append_downloaded_file_content(
        self,
        *,
        session: AgentSessionContext,
        file_content: FileContent,
        url: str,
        sender: Participant | None,
    ) -> None:
        mime_type = file_content.mime_type or "application/octet-stream"
        if mime_type.startswith("image/") and session.supports_images:
            session.append_image_message(
                mime_type=mime_type,
                data=file_content.data,
            )
            return

        if session.supports_files:
            session.append_file_message(
                filename=file_content.name,
                mime_type=mime_type,
                data=file_content.data,
            )
            return

        session.append_user_message(
            self._file_attachment_message(sender=sender, url=url)
        )

    def _append_remote_file_content(
        self,
        *,
        session: AgentSessionContext,
        url: str,
        sender: Participant | None,
    ) -> None:
        guessed_mime_type = self._guess_url_mime_type(url=url)
        if (
            guessed_mime_type is not None
            and guessed_mime_type.startswith("image/")
            and session.supports_images
        ):
            session.append_image_url(url=url)
            return

        if session.supports_files:
            session.append_file_url(url=url)
            return

        session.append_user_message(
            self._file_attachment_message(sender=sender, url=url)
        )

    async def _append_file_content(
        self,
        *,
        session: AgentSessionContext,
        url: str,
        sender: Participant | None,
    ) -> None:
        if urlparse(url).scheme == "room":
            room_path = self._normalize_room_storage_path(url=url)
            file_content = await self._room.storage.download(path=room_path)
            self._append_downloaded_file_content(
                session=session,
                file_content=file_content,
                url=url,
                sender=sender,
            )
            return

        self._append_remote_file_content(
            session=session,
            url=url,
            sender=sender,
        )

    async def _append_turn_content(
        self,
        *,
        session: AgentSessionContext,
        sender: Participant | None,
        turns: list[TurnStart | TurnSteer],
    ) -> None:
        for turn in turns:
            for item in turn.content:
                if isinstance(item, AgentTextContent):
                    session.append_user_message(
                        self._format_live_turn_message(
                            sender=sender,
                            message=item.text,
                        )
                    )
                elif isinstance(item, AgentFileContent):
                    await self._append_file_content(
                        session=session,
                        url=item.url,
                        sender=sender,
                    )

    async def _append_queued_turn_messages(
        self,
        *,
        session: AgentSessionContext,
        queued_messages: list[_QueuedTurnMessage],
    ) -> None:
        for queued_message in queued_messages:
            await self._append_turn_content(
                session=session,
                sender=queued_message.sender,
                turns=[queued_message.request],
            )

    def _turn_error(self, *, message: str, code: str | None = None) -> AgentError:
        return AgentError(message=message, code=code)

    def _sender_for_turn_batch(
        self,
        *,
        queued_messages: list[_QueuedTurnMessage],
    ) -> Participant | None:
        if len(queued_messages) == 0:
            return None

        first_sender = queued_messages[0].sender
        for queued_message in queued_messages[1:]:
            if queued_message.sender != first_sender:
                return None

        return first_sender

    def _turn_ended_rejection(self) -> AgentError:
        return self._turn_error(
            message="turn ended before queued steer was processed",
            code="turn_ended",
        )

    def _drain_queued_turn_messages(
        self,
        *,
        active_turn_queue: asyncio.Queue[_QueuedTurnMessage] | None,
    ) -> list[_QueuedTurnMessage]:
        if active_turn_queue is None:
            return []

        drained_messages: list[_QueuedTurnMessage] = []
        while True:
            try:
                drained_messages.append(active_turn_queue.get_nowait())
            except asyncio.QueueEmpty:
                break

        return drained_messages

    def _drain_pending_turns(self) -> list[_QueuedTurn]:
        drained_turns: list[_QueuedTurn] = []
        if self._priority_turn is not None:
            drained_turns.append(self._priority_turn)
            self._priority_turn = None
        while True:
            try:
                drained_turns.append(self._pending_turns.get_nowait())
            except asyncio.QueueEmpty:
                break

        return drained_turns

    def _has_pending_turns(self) -> bool:
        return self._priority_turn is not None or not self._pending_turns.empty()

    async def _next_pending_turn(self) -> _QueuedTurn:
        if self._priority_turn is not None:
            queued_turn = self._priority_turn
            self._priority_turn = None
            return queued_turn

        return await self._pending_turns.get()

    def _emit_rejected_queued_turn_steers(
        self,
        *,
        queued_messages: list[_QueuedTurnMessage],
        error: AgentError,
    ) -> None:
        for queued_message in queued_messages:
            turn = queued_message.request
            if not isinstance(turn, TurnSteer):
                continue

            self.emit(
                sender=queued_message.sender,
                payload=TurnSteerRejected(
                    type=AGENT_EVENT_TURN_STEER_REJECTED,
                    thread_id=turn.thread_id,
                    turn_id=turn.turn_id,
                    source_message_id=turn.message_id,
                    error=error,
                ),
            )

    def _emit_turn_steered_events(
        self,
        *,
        queued_messages: list[_QueuedTurnMessage],
    ) -> None:
        turn_id = self._turn_id
        thread_id = self.thread_id
        if turn_id is None or thread_id is None:
            return

        for queued_message in queued_messages:
            turn = queued_message.request
            if not isinstance(turn, TurnSteer):
                continue

            self.emit(
                sender=queued_message.sender,
                payload=TurnSteered(
                    type=AGENT_EVENT_TURN_STEERED,
                    thread_id=thread_id,
                    turn_id=turn_id,
                    source_message_id=turn.message_id,
                ),
            )

    async def _apply_pending_turn_steers(
        self,
        *,
        session: AgentSessionContext,
    ) -> bool:
        turn_id = self._turn_id
        if turn_id is None or self._interrupt_requested_turn_id == turn_id:
            return False

        queued_messages = self._drain_queued_turn_messages(
            active_turn_queue=self._active_turn_queue,
        )
        steer_messages = [
            queued_message
            for queued_message in queued_messages
            if isinstance(queued_message.request, TurnSteer)
        ]
        if len(steer_messages) == 0:
            return False

        await self._remove_pending_status_messages(queued_messages=steer_messages)
        self._record_applied_turns(queued_messages=steer_messages)
        await self._append_queued_turn_messages(
            session=session,
            queued_messages=steer_messages,
        )
        self._emit_turn_steered_events(queued_messages=steer_messages)
        return True

    async def _prepare_turn_batch(
        self,
        *,
        queued_messages: list[_QueuedTurnMessage],
        session: AgentSessionContext,
        model: str,
    ) -> tuple[Participant | None, list[Toolkit]]:
        turns = [queued_message.request for queued_message in queued_messages]
        sender = self._sender_for_turn_batch(queued_messages=queued_messages)

        self._record_applied_turns(queued_messages=queued_messages)
        await self._append_queued_turn_messages(
            session=session,
            queued_messages=queued_messages,
        )
        self._emit_turn_steered_events(queued_messages=queued_messages)
        combined_toolkits = await self._build_turn_toolkits(
            model=model,
            turns=turns,
            sender=sender,
        )
        return sender, combined_toolkits

    async def _run_adapter_next(
        self,
        *,
        session: AgentSessionContext,
        sender: Participant | None,
        combined_toolkits: list[Toolkit],
        model: str,
    ) -> None:
        turn_id = self._turn_id
        thread_id = self.thread_id
        if turn_id is None or thread_id is None:
            raise RuntimeError("turn publisher requested without an active turn")

        def publish_event(message: AgentMessage) -> None:
            if self._interrupt_requested_turn_id == turn_id:
                return
            self.emit(sender=sender, payload=message)

        thread_adapter = self.thread_adapter

        def publish_custom_event(event: dict[str, Any]) -> None:
            if self._interrupt_requested_turn_id == turn_id:
                return
            if thread_adapter is None:
                return

            event_type = event.get("type")
            if event_type not in ("agent.event", "codex.event"):
                return

            raw_state = event.get("state")
            if not isinstance(raw_state, str):
                return
            state = raw_state.strip().lower()

            headline = event.get("headline")
            summary = event.get("summary")
            status_text = None
            if isinstance(headline, str) and headline.strip() != "":
                status_text = headline.strip()
            elif isinstance(summary, str) and summary.strip() != "":
                status_text = summary.strip()

            if state in _THREAD_STATUS_ACTIVE_STATES:
                if status_text is None:
                    return
                asyncio.create_task(
                    thread_adapter.set_thread_status(status=status_text)
                )
                return

            if state in _THREAD_STATUS_TERMINAL_STATES:
                asyncio.create_task(thread_adapter.set_thread_status(status="Thinking"))

        handle_event = self.llm_adapter.make_agent_event_publisher(
            turn_id=turn_id,
            thread_id=thread_id,
            callback=publish_event,
            custom_event_callback=publish_custom_event,
        )

        self._active_turn_sender = sender
        had_thread_id = "thread_id" in session.metadata
        previous_thread_id = session.metadata.get("thread_id")
        had_turn_id = "turn_id" in session.metadata
        previous_turn_id = session.metadata.get("turn_id")
        session.metadata["thread_id"] = thread_id
        session.metadata["turn_id"] = turn_id
        try:
            next_task = asyncio.create_task(
                self.llm_adapter.next(
                    context=session,
                    toolkits=combined_toolkits,
                    room=self._room,
                    event_handler=handle_event,
                    steering_callback=lambda: self._apply_pending_turn_steers(
                        session=session
                    ),
                    model=model,
                    on_behalf_of=sender,
                )
            )
            self._active_next_task = next_task
            await next_task
        finally:
            if had_thread_id:
                session.metadata["thread_id"] = previous_thread_id
            else:
                session.metadata.pop("thread_id", None)

            if had_turn_id:
                session.metadata["turn_id"] = previous_turn_id
            else:
                session.metadata.pop("turn_id", None)
            self._active_next_task = None
            self._active_turn_sender = None

    async def _continue_interrupted_turn(
        self,
        *,
        session: AgentSessionContext,
        model: str,
        active_turn_queue: asyncio.Queue[_QueuedTurnMessage],
    ) -> bool:
        queued_messages = self._drain_queued_turn_messages(
            active_turn_queue=active_turn_queue,
        )
        steer_messages = [
            queued_message
            for queued_message in queued_messages
            if isinstance(queued_message.request, TurnSteer)
        ]
        if len(steer_messages) == 0:
            return False

        await self._remove_pending_status_messages(queued_messages=steer_messages)
        self.llm_adapter.on_turn_steer(context=session, interrupted=True)
        sender, combined_toolkits = await self._prepare_turn_batch(
            queued_messages=steer_messages,
            session=session,
            model=model,
        )
        await self._run_adapter_next(
            session=session,
            sender=sender,
            combined_toolkits=combined_toolkits,
            model=model,
        )
        return True

    async def _execute_turn_batch(
        self,
        *,
        queued_messages: list[_QueuedTurnMessage],
        session: AgentSessionContext,
        model: str,
    ) -> None:
        sender, combined_toolkits = await self._prepare_turn_batch(
            queued_messages=queued_messages,
            session=session,
            model=model,
        )
        await self._run_adapter_next(
            session=session,
            sender=sender,
            combined_toolkits=combined_toolkits,
            model=model,
        )

    async def _handle_interrupt(
        self,
        *,
        queued_turn: _QueuedTurn,
        turn_id: str,
        active_turn_queue: asyncio.Queue[_QueuedTurnMessage],
    ) -> Literal["continue", "cancel"] | None:
        if self._interrupt_requested_turn_id != turn_id:
            return None

        interrupt_source_message_id = self._interrupt_source_message_id
        self._interrupt_requested_turn_id = None
        self._interrupt_source_message_id = None
        self._cancel_pending_tool_call_approvals()

        if interrupt_source_message_id is not None:
            self.emit(
                sender=queued_turn.sender,
                payload=TurnInterrupted(
                    type=AGENT_EVENT_TURN_INTERRUPTED,
                    thread_id=queued_turn.request.thread_id,
                    turn_id=turn_id,
                    source_message_id=interrupt_source_message_id,
                ),
            )

        if active_turn_queue.empty():
            return "cancel"
        return "continue"

    async def _run_next_turn(self) -> None:
        queued_turn = await self._next_pending_turn()
        queued_turn_messages = [
            _QueuedTurnMessage(
                sender=queued_turn.sender,
                request=queued_turn.request,
            ),
            *queued_turn.queued_messages,
        ]
        await self._remove_pending_status_messages(queued_messages=queued_turn_messages)
        turn_id = str(uuid.uuid4())
        self._turn_id = turn_id
        self._interrupt_requested_turn_id = None
        self._interrupt_source_message_id = None
        active_turn_queue: asyncio.Queue[_QueuedTurnMessage] = asyncio.Queue()
        self._active_turn_queue = active_turn_queue
        for queued_message in queued_turn_messages:
            active_turn_queue.put_nowait(queued_message)
        self.emit(
            sender=queued_turn.sender,
            payload=TurnStarted(
                type=AGENT_EVENT_TURN_STARTED,
                thread_id=queued_turn.request.thread_id,
                turn_id=turn_id,
                source_message_id=queued_turn.request.message_id,
            ),
        )

        error: AgentError | None = None
        session: AgentSessionContext | None = None
        original_instructions: str | None = None
        try:
            session = await self.ensure_session_context(turn_id=turn_id)
            model = (
                queued_turn.request.model
                if queued_turn.request.model is not None
                else self.llm_adapter.default_model()
            )
            original_instructions = session.instructions
            if queued_turn.request.instructions is not None:
                session.instructions = queued_turn.request.instructions

            continue_interrupted_turn = False
            while True:
                if continue_interrupted_turn:
                    continue_interrupted_turn = False
                    continued = await self._continue_interrupted_turn(
                        session=session,
                        model=model,
                        active_turn_queue=active_turn_queue,
                    )
                    if not continued:
                        error = self._turn_error(
                            message="turn cancelled",
                            code="cancelled",
                        )
                        break
                else:
                    queued_messages = [await active_turn_queue.get()]
                    while True:
                        try:
                            queued_messages.append(active_turn_queue.get_nowait())
                        except asyncio.QueueEmpty:
                            break

                    await self._remove_pending_status_messages(
                        queued_messages=queued_messages
                    )

                    await self._execute_turn_batch(
                        queued_messages=queued_messages,
                        session=session,
                        model=model,
                    )
                interrupt_action = await self._handle_interrupt(
                    queued_turn=queued_turn,
                    turn_id=turn_id,
                    active_turn_queue=active_turn_queue,
                )
                if interrupt_action == "continue":
                    continue_interrupted_turn = True
                    continue
                if interrupt_action == "cancel":
                    error = self._turn_error(message="turn cancelled", code="cancelled")
                    break
                if active_turn_queue.empty():
                    break
        except asyncio.CancelledError:
            interrupt_action = await self._handle_interrupt(
                queued_turn=queued_turn,
                turn_id=turn_id,
                active_turn_queue=active_turn_queue,
            )
            if interrupt_action == "continue":
                continue_interrupted_turn = True
                while True:
                    try:
                        if continue_interrupted_turn:
                            continue_interrupted_turn = False
                            continued = await self._continue_interrupted_turn(
                                session=session,
                                model=model,
                                active_turn_queue=active_turn_queue,
                            )
                            if not continued:
                                error = self._turn_error(
                                    message="turn cancelled",
                                    code="cancelled",
                                )
                                break
                        else:
                            queued_messages = [await active_turn_queue.get()]
                            while True:
                                try:
                                    queued_messages.append(
                                        active_turn_queue.get_nowait()
                                    )
                                except asyncio.QueueEmpty:
                                    break

                            await self._remove_pending_status_messages(
                                queued_messages=queued_messages
                            )
                            await self._execute_turn_batch(
                                queued_messages=queued_messages,
                                session=session,
                                model=model,
                            )
                    except asyncio.CancelledError:
                        interrupt_action = await self._handle_interrupt(
                            queued_turn=queued_turn,
                            turn_id=turn_id,
                            active_turn_queue=active_turn_queue,
                        )
                        if interrupt_action == "continue":
                            continue_interrupted_turn = True
                            continue
                        error = self._turn_error(
                            message="turn cancelled",
                            code="cancelled",
                        )
                        break

                    interrupt_action = await self._handle_interrupt(
                        queued_turn=queued_turn,
                        turn_id=turn_id,
                        active_turn_queue=active_turn_queue,
                    )
                    if interrupt_action == "continue":
                        continue_interrupted_turn = True
                        continue
                    if interrupt_action == "cancel":
                        error = self._turn_error(
                            message="turn cancelled",
                            code="cancelled",
                        )
                        break
                    if active_turn_queue.empty():
                        break
            else:
                error = self._turn_error(message="turn cancelled", code="cancelled")
        except Exception as exc:
            logger.exception("turn failed")
            error_message = str(exc) if str(exc) != "" else exc.__class__.__name__
            error = self._turn_error(
                message=error_message,
                code=exc.__class__.__name__,
            )
        finally:
            self._interrupt_requested_turn_id = None
            self._interrupt_source_message_id = None
            self._active_turn_queue = None
            remaining_queued_messages = self._drain_queued_turn_messages(
                active_turn_queue=active_turn_queue,
            )
            if len(remaining_queued_messages) > 0:
                await self._remove_pending_status_messages(
                    queued_messages=remaining_queued_messages
                )
                self._emit_rejected_queued_turn_steers(
                    queued_messages=remaining_queued_messages,
                    error=self._turn_ended_rejection(),
                )
            self._cancel_pending_tool_call_approvals()
            if session is not None:
                session.instructions = original_instructions
            self.emit(
                sender=queued_turn.sender,
                payload=TurnEnded(
                    type=AGENT_EVENT_TURN_ENDED,
                    thread_id=queued_turn.request.thread_id,
                    turn_id=turn_id,
                    error=error,
                ),
            )

    def _schedule_next_turn(self) -> None:
        if (
            self._turn_task is not None
            or self._stop.is_set()
            or not self._has_pending_turns()
        ):
            return

        self._turn_task = asyncio.create_task(self._run_next_turn())
        self._turn_task.add_done_callback(self._on_turn_done)

    def _on_turn_done(self, task: asyncio.Task[None]) -> None:
        if task.cancelled():
            logger.debug("turn task cancelled")
        else:
            exc = task.exception()
            if exc is not None:
                logger.exception("turn failed", exc_info=exc)

        self._turn_id = None
        self._turn_task = None
        self._active_turn_queue = None

        if not self._stop.is_set():
            self._schedule_next_turn()

    async def on_turn_start(self, message: Message) -> None:
        turn = _coerce_message_data(message.data, TurnStart)
        if turn.instructions is None and self._turn_instructions_provider is not None:
            instructions = await self._turn_instructions_provider(message.sender)
            if instructions is not None:
                turn = turn.model_copy(update={"instructions": instructions})
        should_track_pending_status = (
            self._turn_task is not None or self._has_pending_turns()
        )
        await self._pending_turns.put(
            _QueuedTurn(
                sender=message.sender,
                request=turn,
            )
        )
        self.emit(
            sender=message.sender,
            payload=TurnStartAccepted(
                type=AGENT_EVENT_TURN_START_ACCEPTED,
                thread_id=turn.thread_id,
                source_message_id=turn.message_id,
            ),
        )
        if should_track_pending_status:
            await self._add_pending_status_messages(
                queued_messages=[
                    _QueuedTurnMessage(
                        sender=message.sender,
                        request=turn,
                    )
                ]
            )
        self._schedule_next_turn()

    async def on_turn_steer(self, message: Message) -> None:
        turn = _coerce_message_data(message.data, TurnSteer)

        active_turn_queue = self._active_turn_queue
        if (
            self._turn_id is None
            or self._turn_id != turn.turn_id
            or active_turn_queue is None
        ):
            rejection = self._turn_error(
                message="turn is not in progress",
                code="turn_not_in_progress",
            )
            self.emit(
                sender=message.sender,
                payload=TurnSteerRejected(
                    type=AGENT_EVENT_TURN_STEER_REJECTED,
                    thread_id=turn.thread_id,
                    turn_id=turn.turn_id,
                    source_message_id=turn.message_id,
                    error=rejection,
                ),
            )
            return

        active_turn_queue.put_nowait(
            _QueuedTurnMessage(
                sender=message.sender,
                request=turn,
            )
        )
        await self._add_pending_status_messages(
            queued_messages=[
                _QueuedTurnMessage(
                    sender=message.sender,
                    request=turn,
                )
            ]
        )
        self.emit(
            sender=message.sender,
            payload=TurnSteerAccepted(
                type=AGENT_EVENT_TURN_STEER_ACCEPTED,
                thread_id=turn.thread_id,
                turn_id=turn.turn_id,
                source_message_id=turn.message_id,
            ),
        )

    async def on_turn_interrupt(self, message: Message) -> None:
        turn = _coerce_message_data(message.data, TurnInterrupt)
        if self._turn_id != turn.turn_id or self._turn_task is None:
            return

        self._interrupt_requested_turn_id = turn.turn_id
        self._interrupt_source_message_id = turn.message_id
        self.emit(
            sender=message.sender,
            payload=TurnInterruptAccepted(
                type=AGENT_EVENT_TURN_INTERRUPT_ACCEPTED,
                thread_id=turn.thread_id,
                turn_id=turn.turn_id,
                source_message_id=turn.message_id,
            ),
        )
        active_next_task = self._active_next_task
        if active_next_task is not None:
            active_next_task.cancel()

    async def on_clear_thread(self, message: Message) -> None:
        clear_thread = _coerce_message_data(message.data, ClearThread)
        if self.thread_id != clear_thread.thread_id:
            return

        turn_task = self._turn_task
        if turn_task is not None:
            active_next_task = self._active_next_task
            if active_next_task is not None:
                active_next_task.cancel()
            turn_task.cancel()
            with contextlib.suppress(asyncio.CancelledError):
                await turn_task

        self._active_turn_queue = None
        self._active_turn_sender = None
        self._turn_id = None
        self._interrupt_requested_turn_id = None
        self._interrupt_source_message_id = None
        self._drain_pending_turns()
        self._cancel_pending_tool_call_approvals()
        await self._clear_pending_status_messages()

        if self._session_context is not None:
            await self._session_context.close()
            self._session_context = None

        thread_adapter = self.thread_adapter
        if thread_adapter is not None:
            await thread_adapter.clear_thread()

        self.emit(
            sender=message.sender,
            payload=ThreadCleared(
                type=AGENT_EVENT_THREAD_CLEARED,
                thread_id=clear_thread.thread_id,
                source_message_id=clear_thread.message_id,
            ),
        )

    async def on_tool_call_approve(self, message: Message) -> None:
        approval = _coerce_message_data(message.data, ApproveAgentToolCall)
        if self._turn_id != approval.turn_id:
            return

        await self._resolve_tool_call_approval(
            item_id=approval.item_id,
            approved=True,
        )

    async def on_tool_call_reject(self, message: Message) -> None:
        rejection = _coerce_message_data(message.data, RejectAgentToolCall)
        if self._turn_id != rejection.turn_id:
            return

        await self._resolve_tool_call_approval(
            item_id=rejection.item_id,
            approved=False,
        )

    async def on_stop(self) -> None:
        if self._turn_task is not None:
            active_next_task = self._active_next_task
            if active_next_task is not None:
                active_next_task.cancel()
            self._turn_task.cancel()
            with contextlib.suppress(asyncio.CancelledError):
                await self._turn_task
        self._active_turn_queue = None
        self._active_turn_sender = None
        self._interrupt_requested_turn_id = None
        self._interrupt_source_message_id = None
        self._cancel_pending_tool_call_approvals()
        await self._clear_pending_status_messages()

        if self._session_context is not None:
            await self._session_context.close()
            self._session_context = None

    async def on_message(self, message: Message) -> None:
        handler = self._handlers.get(message.data.type)
        if handler is None:
            return

        try:
            await handler(message)
        except Exception:
            logger.exception("llm agent process failed to handle message")



from __future__ import annotations

import asyncio
import base64
import contextlib
from dataclasses import dataclass, field
from email import message_from_bytes
import email.utils
import json
import logging
import os
import posixpath
import re
from typing import Any
import uuid

from meshagent.api import Participant, RoomClient, RoomException
from meshagent.api.room_server_client import TextDataType
from meshagent.tools import FileContent, FunctionTool, ToolContext, Toolkit
from email.policy import default

from .adapter import LLMAdapter
from .mail_common import (
    SmtpConfiguration,
    create_email_message,
    create_reply_email_message,
    iter_message_attachments,
    message_to_json,
    should_reply_to_message,
)
from .messages import (
    AGENT_EVENT_TEXT_CONTENT_DELTA,
    AGENT_EVENT_TEXT_CONTENT_ENDED,
    AGENT_EVENT_TEXT_CONTENT_STARTED,
    AGENT_EVENT_THREAD_CLEARED,
    AGENT_EVENT_TURN_ENDED,
    AGENT_EVENT_TURN_STARTED,
    AGENT_MESSAGE_TURN_START,
    AgentFileContent,
    AgentTextContent,
    AgentTextContentDelta,
    AgentTextContentEnded,
    AgentTextContentStarted,
    ThreadCleared,
    TurnEnded,
    TurnStart,
    TurnStarted,
)
from .process import Message
from .threaded_channel import ThreadedChannel

logger = logging.getLogger("mail-channel")


class _MissingAioSmtplib:
    async def send(self, *args: Any, **kwargs: Any) -> None:
        del args
        del kwargs
        raise ModuleNotFoundError(
            "aiosmtplib is required to use MailChannel SMTP features"
        )


try:
    import aiosmtplib
except ModuleNotFoundError:
    aiosmtplib = _MissingAioSmtplib()

_MAIL_TABLE_NAME = "emails"
_MAIL_STORAGE_ROOT = ".emails"


@dataclass(slots=True)
class _PendingInboundMail:
    thread_id: str
    message: dict[str, Any]


@dataclass(slots=True)
class _ActiveMailTurn:
    thread_id: str
    source_message: dict[str, Any]
    text_by_item_id: dict[str, str] = field(default_factory=dict)
    completed_text_parts: list[str] = field(default_factory=list)


class MailChannel(ThreadedChannel):
    def __init__(
        self,
        *,
        room: RoomClient,
        queue_name: str,
        email_address: str,
        threading_mode: str | None = None,
        thread_dir: str | None = None,
        llm_adapter: LLMAdapter | None = None,
        domain: str = os.getenv("MESHAGENT_MAIL_DOMAIN", "mail.meshagent.com"),
        smtp: SmtpConfiguration | None = None,
        whitelist: list[str] | None = None,
        reply_all: bool = False,
        enable_attachments: bool = True,
    ) -> None:
        super().__init__(
            room=room,
            threading_mode=threading_mode,
            thread_dir=thread_dir,
            llm_adapter=llm_adapter,
        )
        normalized_queue = queue_name.strip()
        if normalized_queue == "":
            raise ValueError("queue_name must not be empty")

        normalized_email = email_address.strip()
        if normalized_email == "":
            raise ValueError("email_address must not be empty")

        self._queue_name = normalized_queue
        self._email_address = normalized_email
        self._domain = domain
        self._smtp = smtp if smtp is not None else SmtpConfiguration()
        self._whitelist = list(whitelist) if whitelist is not None else None
        self._reply_all = reply_all
        self._enable_attachments = enable_attachments
        self._receive_task: asyncio.Task[None] | None = None
        self._pending_messages_by_source_id: dict[str, _PendingInboundMail] = {}
        self._active_turns_by_turn_id: dict[str, _ActiveMailTurn] = {}
        self._database_namespace_supported: bool = True

    def _default_thread_dir_fallback_name(self) -> str:
        return "mail"

    def _default_thread_name(self) -> str:
        return "New Mail"

    def handles(self, message: Message) -> bool:
        return message.data.type in {
            AGENT_EVENT_TEXT_CONTENT_STARTED,
            AGENT_EVENT_TEXT_CONTENT_DELTA,
            AGENT_EVENT_TEXT_CONTENT_ENDED,
            AGENT_EVENT_TURN_STARTED,
            AGENT_EVENT_TURN_ENDED,
            AGENT_EVENT_THREAD_CLEARED,
        }

    def get_agent_toolkits(self) -> list[Toolkit]:
        return [Toolkit(name="mail", tools=[self._make_new_email_thread_tool()])]

    async def on_start(self) -> None:
        await self.publish_thread_attributes()
        await self.open_thread_list_document()
        await self._ensure_database_table()
        self._receive_task = asyncio.create_task(self._receive_loop())

    async def on_stop(self) -> None:
        receive_task = self._receive_task
        self._receive_task = None
        if receive_task is not None:
            receive_task.cancel()
            with contextlib.suppress(asyncio.CancelledError):
                await receive_task
        await self.close_thread_list_document()
        self._pending_messages_by_source_id.clear()
        self._active_turns_by_turn_id.clear()

    async def on_message(self, message: Message) -> None:
        data = message.data
        if isinstance(data, TurnStarted):
            pending = self._pending_messages_by_source_id.pop(
                data.source_message_id,
                None,
            )
            if pending is None:
                return
            self._active_turns_by_turn_id[data.turn_id] = _ActiveMailTurn(
                thread_id=pending.thread_id,
                source_message=pending.message,
            )
            return

        if isinstance(data, AgentTextContentStarted):
            active = self._active_turns_by_turn_id.get(data.turn_id)
            if active is None:
                return
            active.text_by_item_id[data.item_id] = ""
            return

        if isinstance(data, AgentTextContentDelta):
            active = self._active_turns_by_turn_id.get(data.turn_id)
            if active is None:
                return
            current_text = active.text_by_item_id.get(data.item_id, "")
            active.text_by_item_id[data.item_id] = current_text + data.text
            return

        if isinstance(data, AgentTextContentEnded):
            active = self._active_turns_by_turn_id.get(data.turn_id)
            if active is None:
                return
            completed_text = active.text_by_item_id.pop(data.item_id, "")
            if completed_text.strip() != "":
                active.completed_text_parts.append(completed_text)
            return

        if isinstance(data, ThreadCleared):
            self._clear_thread_state(thread_id=data.thread_id)
            return

        if not isinstance(data, TurnEnded):
            return

        active = self._active_turns_by_turn_id.pop(data.turn_id, None)
        if active is None:
            return

        if data.error is not None:
            return

        reply = self._reply_text_for_turn(active=active)
        if reply == "":
            return

        try:
            await self.send_reply_message(
                thread_id=active.thread_id,
                source_message=active.source_message,
                reply=reply,
            )
        except Exception:
            logger.exception(
                "failed sending email reply for thread %s", active.thread_id
            )

    async def start_thread(
        self,
        *,
        thread_id: str,
        to_address: str,
        subject: str,
        body: str,
        attachments: list[FileContent] | None = None,
    ) -> dict[str, Any]:
        message = create_email_message(
            to_address=to_address,
            from_address=self._email_address,
            subject=subject,
            body=body,
        )
        for attachment in attachments or []:
            maintype, subtype = attachment.mime_type.split("/", 1)
            message.add_attachment(
                attachment.data,
                maintype=maintype,
                subtype=subtype,
                filename=attachment.name,
            )

        saved_message = await self._save_email_message(
            content=message.as_bytes(),
            role="agent",
            thread_id=thread_id,
        )
        await self._smtp_send(message=message)
        return saved_message

    async def send_reply_message(
        self,
        *,
        thread_id: str,
        source_message: dict[str, Any],
        reply: str,
    ) -> dict[str, Any]:
        message = create_reply_email_message(
            message=source_message,
            from_address=self._email_address,
            body=reply,
            email_address=self._email_address,
            reply_all=self._reply_all,
        )
        saved_message = await self._save_email_message(
            content=message.as_bytes(),
            role="agent",
            thread_id=thread_id,
        )
        await self._smtp_send(message=message)
        return saved_message

    async def _receive_loop(self) -> None:
        while not self._stop.is_set():
            try:
                queued_message = await self._room.queues.receive(
                    name=self._queue_name,
                    create=True,
                    wait=True,
                )
            except asyncio.CancelledError:
                raise
            except Exception:
                logger.exception(
                    "mail queue receive failed for queue %s",
                    self._queue_name,
                )
                await asyncio.sleep(1)
                continue

            if queued_message is None:
                continue

            try:
                await self._handle_inbound_queue_message(queued_message=queued_message)
            except asyncio.CancelledError:
                raise
            except Exception:
                logger.exception(
                    "mail channel failed to process inbound queue message for %s",
                    self._queue_name,
                )

    async def _handle_inbound_queue_message(self, *, queued_message: Any) -> None:
        if not isinstance(queued_message, dict):
            logger.warning(
                "ignoring mail queue payload with unsupported type %s",
                type(queued_message),
            )
            return

        encoded_message = queued_message.get("base64")
        if not isinstance(encoded_message, str) or encoded_message.strip() == "":
            logger.warning("ignoring mail queue payload without base64 email content")
            return

        raw_bytes = base64.b64decode(encoded_message)
        email_message = message_from_bytes(raw_bytes, policy=default)
        parsed_message = message_to_json(message=email_message, role="user")
        attachment_names = [
            file_name for file_name, _ in iter_message_attachments(email_message)
        ]

        should_reply, reason = should_reply_to_message(
            message=parsed_message,
            email_address=self._email_address,
            whitelist=self._whitelist,
        )
        if not should_reply:
            logger.info(
                "discarding inbound mail for %s: %s",
                self._queue_name,
                reason or "message is not replyable",
            )
            return

        thread_id = await self._resolve_thread_id(
            message=parsed_message,
            attachments=attachment_names,
        )
        self._record_inbound_thread_activity(
            thread_id=thread_id,
            message=parsed_message,
            attachments=attachment_names,
        )
        saved_message = await self._save_email_message(
            content=raw_bytes,
            role="user",
            thread_id=thread_id,
        )

        turn_start = TurnStart(
            type=AGENT_MESSAGE_TURN_START,
            thread_id=thread_id,
            content=self._content_from_mail_message(message=saved_message),
        )
        self._pending_messages_by_source_id[turn_start.message_id] = (
            _PendingInboundMail(
                thread_id=thread_id,
                message=saved_message,
            )
        )
        self.emit(
            sender=self._participant_for_message(message=saved_message),
            payload=turn_start,
        )

    async def _ensure_database_table(self) -> None:
        table_name, namespace = self._database_table_ref()
        try:
            await self._room.database.create_table_with_schema(
                name=table_name,
                schema={
                    "id": TextDataType(nullable=False),
                    "thread_id": TextDataType(nullable=False),
                    "in_reply_to": TextDataType(),
                    "role": TextDataType(nullable=False),
                    "json": TextDataType(nullable=False),
                },
                mode="create_if_not_exists",
                namespace=namespace,
            )
        except RoomException as exc:
            if (
                namespace is None
                or not self._is_namespace_location_error(exc=exc)
                or not self._database_namespace_supported
            ):
                raise

            self._database_namespace_supported = False
            fallback_table_name, fallback_namespace = self._database_table_ref()
            logger.warning(
                "mail channel falling back to flat database table %s because the room server rejected namespaced table creation",
                fallback_table_name,
            )
            await self._room.database.create_table_with_schema(
                name=fallback_table_name,
                schema={
                    "id": TextDataType(nullable=False),
                    "thread_id": TextDataType(nullable=False),
                    "in_reply_to": TextDataType(),
                    "role": TextDataType(nullable=False),
                    "json": TextDataType(nullable=False),
                },
                mode="create_if_not_exists",
                namespace=fallback_namespace,
            )

    async def _load_message(self, *, message_id: str) -> dict[str, Any] | None:
        table_name, namespace = self._database_table_ref()
        results = await self._room.database.search(
            table=table_name,
            where={"id": message_id},
            namespace=namespace,
        )
        if len(results) == 0:
            return None
        raw_json = results[0].get("json")
        if not isinstance(raw_json, str):
            return None
        return json.loads(raw_json)

    async def _resolve_thread_id(
        self,
        *,
        message: dict[str, Any],
        attachments: list[str] | None = None,
    ) -> str:
        in_reply_to = message.get("in_reply_to")
        if isinstance(in_reply_to, str) and in_reply_to.strip() != "":
            parent_message = await self._load_message(message_id=in_reply_to.strip())
            if parent_message is not None:
                thread_id = parent_message.get("thread_id")
                if isinstance(thread_id, str) and thread_id.strip() != "":
                    return thread_id

        path, _ = await self.new_thread(
            message_text=self._thread_name_message_text(message=message),
            attachments=attachments,
        )
        return path

    def _thread_name_message_text(self, *, message: dict[str, Any]) -> str:
        subject = str(message.get("subject") or "").strip()
        body = str(message.get("body") or "").strip()
        sender = str(message.get("from") or "").strip()
        return subject or body or sender

    async def _save_email_message(
        self,
        *,
        content: bytes,
        role: str,
        thread_id: str,
    ) -> dict[str, Any]:
        email_message = message_from_bytes(content, policy=default)
        queued_message = message_to_json(message=email_message, role=role)
        queued_message["thread_id"] = thread_id

        folder_path = posixpath.join(
            uuid.uuid4().hex[:8],
            uuid.uuid4().hex[:8],
        )
        base_path = posixpath.join(_MAIL_STORAGE_ROOT, folder_path)
        queued_message["path"] = posixpath.join(base_path, "message.json")

        for file_name, attachment_bytes in iter_message_attachments(email_message):
            attachment_path = posixpath.join(base_path, "attachments", file_name)
            await self._room.storage.upload(path=attachment_path, data=attachment_bytes)
            queued_message["attachments"].append(attachment_path)

        await self._room.storage.upload(
            path=posixpath.join(base_path, "message.eml"),
            data=content,
        )
        await self._room.storage.upload(
            path=posixpath.join(base_path, "message.json"),
            data=json.dumps(queued_message, indent=4).encode("utf-8"),
        )
        table_name, namespace = self._database_table_ref()
        await self._room.database.insert(
            table=table_name,
            namespace=namespace,
            records=[
                {
                    "id": queued_message["id"],
                    "thread_id": thread_id,
                    "in_reply_to": queued_message.get("in_reply_to"),
                    "role": role,
                    "json": json.dumps(queued_message),
                }
            ],
        )
        return queued_message

    def _content_from_mail_message(
        self,
        *,
        message: dict[str, Any],
    ) -> list[AgentTextContent | AgentFileContent]:
        body = str(message.get("body") or "")
        from_header = str(message.get("from") or "")
        subject = str(message.get("subject") or "")
        to_header = ", ".join(str(item) for item in message.get("to") or [])
        cc_header = ", ".join(str(item) for item in message.get("cc") or [])

        text_lines = ["Incoming email message:"]
        if from_header != "":
            text_lines.append(f"From: {from_header}")
        if to_header != "":
            text_lines.append(f"To: {to_header}")
        if cc_header != "":
            text_lines.append(f"Cc: {cc_header}")
        if subject != "":
            text_lines.append(f"Subject: {subject}")
        if body != "":
            text_lines.extend(["", body])

        content: list[AgentTextContent | AgentFileContent] = [
            AgentTextContent(
                type="text",
                text="\n".join(text_lines),
            )
        ]
        if self._enable_attachments:
            for attachment_path in message.get("attachments") or []:
                if not isinstance(attachment_path, str):
                    continue
                normalized = attachment_path.strip().lstrip("/")
                if normalized == "":
                    continue
                content.append(
                    AgentFileContent(
                        type="file",
                        url=f"room:///{normalized}",
                    )
                )

        return content

    def _participant_for_message(self, *, message: dict[str, Any]) -> Participant:
        from_header = str(message.get("from") or "")
        display_name, address = email.utils.parseaddr(from_header)
        participant_name = display_name.strip() or address.strip() or "email user"
        participant_id = f"mail:{(address or message.get('id') or uuid.uuid4().hex)}"
        return Participant(id=participant_id, attributes={"name": participant_name})

    def _reply_text_for_turn(self, *, active: _ActiveMailTurn) -> str:
        parts = [part for part in active.completed_text_parts if part.strip() != ""]
        for text in active.text_by_item_id.values():
            if text.strip() != "":
                parts.append(text)
        return "\n\n".join(parts).strip()

    def _clear_thread_state(self, *, thread_id: str) -> None:
        self._pending_messages_by_source_id = {
            source_id: pending
            for source_id, pending in self._pending_messages_by_source_id.items()
            if pending.thread_id != thread_id
        }
        self._active_turns_by_turn_id = {
            turn_id: active
            for turn_id, active in self._active_turns_by_turn_id.items()
            if active.thread_id != thread_id
        }

    async def _smtp_send(self, *, message) -> None:
        username = self._smtp.username
        if username is None:
            username_value = self._room.local_participant.get_attribute("name")
            username = username_value if isinstance(username_value, str) else None

        password = self._smtp.password
        if password is None:
            password = self._room.protocol.token

        hostname = (
            self._smtp.hostname if self._smtp.hostname is not None else self._domain
        )
        await aiosmtplib.send(
            message,
            hostname=hostname,
            port=self._smtp.port,
            username=username,
            password=password,
            local_hostname=self._smtp.effective_local_hostname(),
        )

    def _record_inbound_thread_activity(
        self,
        *,
        thread_id: str,
        message: dict[str, Any],
        attachments: list[str] | None = None,
    ) -> None:
        self.bump_thread(
            path=thread_id,
            name=self.fallback_thread_name(
                message_text=self._thread_name_message_text(message=message),
                attachments=attachments,
            ),
        )

    def _database_namespace(self) -> list[str]:
        return [part for part in self.get_thread_dir().split("/") if part != ""]

    @staticmethod
    def _is_namespace_location_error(*, exc: RoomException) -> bool:
        return "Location must be provided when namespace is not empty" in str(exc)

    @staticmethod
    def _sanitize_table_name_component(*, value: str) -> str:
        normalized = re.sub(r"[^A-Za-z0-9_]+", "_", value.strip())
        normalized = normalized.strip("_")
        if normalized == "":
            return "root"
        return normalized

    def _flat_mail_table_name(self) -> str:
        namespace = self._database_namespace()
        if len(namespace) == 0:
            return _MAIL_TABLE_NAME

        suffix = "__".join(
            self._sanitize_table_name_component(value=part) for part in namespace
        )
        return f"{_MAIL_TABLE_NAME}__{suffix}"

    def _database_table_ref(self) -> tuple[str, list[str] | None]:
        namespace = self._database_namespace()
        if self._database_namespace_supported:
            return _MAIL_TABLE_NAME, namespace
        return self._flat_mail_table_name(), None

    def _thread_id_from_tool_context(self, *, context: ToolContext) -> str:
        caller_context = context.caller_context
        if not isinstance(caller_context, dict):
            raise RoomException("mail tool requires a thread_id in caller_context")

        thread_id = caller_context.get("thread_id")
        if not isinstance(thread_id, str) or thread_id.strip() == "":
            raise RoomException("mail tool requires a non-empty thread_id")
        return thread_id

    async def _attachment_files_from_paths(
        self,
        *,
        context: ToolContext,
        attachments: list[str],
    ) -> list[FileContent]:
        files: list[FileContent] = []
        for attachment_path in attachments:
            try:
                files.append(await context.room.storage.download(path=attachment_path))
            except Exception as exc:
                logger.error(
                    "unable to download file %s",
                    attachment_path,
                    exc_info=exc,
                )
                raise RoomException(
                    f"Could not download a file from the room with the path {attachment_path}. Are you sure the path is correct file?"
                ) from exc
        return files

    def _make_new_email_thread_tool(self) -> FunctionTool:
        outer = self

        class NewEmailThreadTool(FunctionTool):
            def __init__(self) -> None:
                super().__init__(
                    name="new_email_thread",
                    description="starts a new outbound email thread from the current agent thread",
                    supports_context=True,
                    input_schema={
                        "type": "object",
                        "additionalProperties": False,
                        "required": ["to", "subject", "body", "attachments"],
                        "properties": {
                            "to": {"type": "string"},
                            "subject": {"type": "string"},
                            "body": {"type": "string"},
                            "attachments": {
                                "type": "array",
                                "items": {"type": "string"},
                                "description": "a list of room storage paths to attach",
                            },
                        },
                    },
                )

            async def execute(
                self,
                context: ToolContext,
                *,
                to: str,
                subject: str,
                body: str,
                attachments: list[str],
            ) -> dict[str, Any]:
                thread_id = outer._thread_id_from_tool_context(context=context)
                files = await outer._attachment_files_from_paths(
                    context=context,
                    attachments=attachments,
                )
                await outer.start_thread(
                    thread_id=thread_id,
                    to_address=to,
                    subject=subject,
                    body=body,
                    attachments=files,
                )
                return {}

        return NewEmailThreadTool()



from pydantic import BaseModel
from typing import Optional


class RulesConfig(BaseModel):
    # Rules that apply for every client
    rules: Optional[list[str]] = None

    # Rules that only apply when participant's client attribute matches the given key
    client_rules: Optional[dict[str, list[str]]] = None

    @staticmethod
    def parse(text: str):
        rules = []
        client_rules = {}

        client = None
        for line in text.splitlines():
            if line.startswith("#"):
                continue

            line = line.strip()
            if len(line) == 0:
                continue

            if line.startswith("[") and line.endswith("]"):
                client = line.strip("[]")
                client_rules[client] = []
            else:
                if client is None:
                    rules.append(line)
                else:
                    client_rules[client].append(line)

        return RulesConfig(
            rules=rules,
            client_rules=client_rules,
        )



from .agent import (
    Agent,
    AgentSessionContext,
    RequiredToolkit,
    SingleRoomAgent,
)

from .context import TaskContext
from .task_runner import TaskRunner
from .threaded_task_runner import ThreadedTaskRunner, ThreadingMode
from .development import connect_development_agent
from .listener import Listener, ListenerContext
from .adapter import ToolResponseAdapter, LLMAdapter
from .image_captioner import ImageCaptioner, LLMImageCaptioner
from .process_thread_adapter import AgentProcessThreadAdapter
from .threaded_channel import ThreadedChannel
from .chat_channel import ChatChannel
from .legacy_chat_channel import LegacyChatChannel
from .mail_channel import MailChannel
from .queue_channel import QueueChannel
from .toolkit_channel import ToolkitChannel
from .thread_schema import thread_schema, thread_list_schema
from .version import __version__


__all__ = [
    Agent,
    TaskContext,
    AgentSessionContext,
    RequiredToolkit,
    TaskRunner,
    ThreadedTaskRunner,
    ThreadingMode,
    SingleRoomAgent,
    connect_development_agent,
    Listener,
    ListenerContext,
    ToolResponseAdapter,
    LLMAdapter,
    ImageCaptioner,
    LLMImageCaptioner,
    AgentProcessThreadAdapter,
    ThreadedChannel,
    ChatChannel,
    LegacyChatChannel,
    MailChannel,
    QueueChannel,
    ToolkitChannel,
    thread_schema,
    thread_list_schema,
    __version__,
]
